"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
import time
from copy import deepcopy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Optional, Sequence

import numpy as np
import torch
import torch.distributed.checkpoint as dcp
from omegaconf import OmegaConf
from torch.distributed.checkpoint.format_utils import dcp_to_torch_save
from torch.distributed.checkpoint.state_dict import (
    get_model_state_dict,
    get_state_dict,
    set_model_state_dict,
    set_state_dict,
)
from torch.distributed.checkpoint.stateful import Stateful
from torch.distributed.device_mesh import init_device_mesh
from torch.distributed.fsdp import (
    FullyShardedDataParallel,
    ShardedStateDictConfig,
    ShardingStrategy,
    StateDictType,
)
from torch.distributed.fsdp.wrap import ModuleWrapPolicy
from torch.profiler import record_function
from torchtnt.framework import EvalUnit, State, TrainUnit
from torchtnt.utils.prepare_module import prepare_module

from fairchem.core.common import distutils, gp_utils
from fairchem.core.common.distutils import (
    get_device_for_local_rank,
)
from fairchem.core.common.logger import WandBSingletonLogger
from fairchem.core.common.registry import registry
from fairchem.core.common.utils import StrEnum
from fairchem.core.components.train.train_runner import Checkpointable
from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.datasets.collaters.mt_collater import MTCollater
from fairchem.core.modules.normalization.element_references import (  # noqa: TCH001
    ElementReferences,
)
from fairchem.core.modules.normalization.normalizer import Normalizer  # noqa: TCH001
from fairchem.core.modules.scheduler import CosineLRLambda
from fairchem.core.units.mlip_unit._metrics import Metrics, get_metrics_fn
from fairchem.core.units.mlip_unit.api.inference import (
    MLIPInferenceCheckpoint,
)
from fairchem.core.units.mlip_unit.utils import load_inference_model

if TYPE_CHECKING:
    from omegaconf import DictConfig

# this is a config generated on the fly and can be used to resume a run for a given checkpoint
UNIT_RESUME_CONFIG = "resume.yaml"

# this represents the inference only checkpoint generated at each checkpoint
UNIT_INFERENCE_CHECKPOINT = "inference_ckpt.pt"


@dataclass
class OutputSpec:
    dim: list[int]
    dtype: str


class TrainStrategy(StrEnum):
    DDP = "ddp"
    FSDP = "fsdp"


@dataclass
class Task:
    name: str
    level: str
    property: str
    out_spec: OutputSpec
    normalizer: Normalizer
    datasets: list[str]
    loss_fn: torch.nn.Module | None = None
    element_references: Optional[ElementReferences] = None
    metrics: list[str] = field(default_factory=list)
    train_on_free_atoms: bool = True
    eval_on_free_atoms: bool = True
    inference_only: bool = False


DEFAULT_EXCLUDE_KEYS = [
    "id",  # only oc20,oc22 have this
    "fid",  # only oc20,oc22 have this
    "absolute_idx",  # only ani has this
    "target_pos",  # only ani has this
    "ref_energy",  # only ani/geom have this
    "pbc",  # only ani/transition1x have this
    "nads",  # oc22
    "oc22",  # oc22
    "formation_energy",  # spice
    "total_charge",  # spice
]


def filter_inference_only_tasks(tasks: Sequence[Task]) -> list[Task]:
    """Filter out tasks that are marked as inference_only."""
    return [task for task in tasks if not task.inference_only]


def convert_train_checkpoint_to_inference_checkpoint(
    dcp_checkpoint_loc: str, checkpoint_loc: str
) -> None:
    dcp_to_torch_save(dcp_checkpoint_loc, checkpoint_loc)

    inference_ckpt = torch.load(
        checkpoint_loc, map_location="cpu", weights_only=False
    )  # DCP model config
    train_eval_unit_state = inference_ckpt["config"]["runner"]["train_eval_unit"]
    unit_state = inference_ckpt["unit_state"]
    torch.save(
        MLIPInferenceCheckpoint(
            model_state_dict=unit_state["model"],
            ema_state_dict=unit_state["ema"],
            model_config=train_eval_unit_state["model"],
            tasks_config=train_eval_unit_state["tasks"],
        ),
        checkpoint_loc,
    )


def initialize_finetuning_model(
    checkpoint_location: str,
    overrides: dict | None = None,
    heads: dict | None = None,
    strict: bool = True,
) -> torch.nn.Module:
    model, checkpoint = load_inference_model(
        checkpoint_location, overrides, strict=strict
    )

    logging.warning(
        f"initialize_finetuning_model starting from checkpoint_location: {checkpoint_location}"
    )

    # if no heads are provided use heads from the checkpoint
    if heads is None:
        model.finetune_model_full_config = checkpoint.model_config
        return model

    checkpoint.model_config["heads"] = deepcopy(heads)
    model.finetune_model_full_config = checkpoint.model_config

    model.output_heads = None
    model.heads = heads
    del model.output_heads
    model.output_heads = {}
    head_names_sorted = sorted(heads.keys())
    assert len(set(head_names_sorted)) == len(
        head_names_sorted
    ), "Head names must be unique!"
    for head_name in head_names_sorted:
        head_config = heads[head_name]
        if "module" not in head_config:
            raise ValueError(
                f"{head_name} head does not specify module to use for the head"
            )
        module_name = head_config.pop("module")
        model.output_heads[head_name] = registry.get_model_class(module_name)(
            model.backbone,
            **head_config,
        )
    model.output_heads = torch.nn.ModuleDict(model.output_heads)
    return model


def get_output_mask(batch: AtomicData, task: Task) -> dict[str, torch.Tensor]:
    """Get a dictionary of boolean masks for each task and dataset in a batch.

    Comment(@abhshkdz): Structures in our `batch` are a mix from various
    sources, e.g. OC20, OC22, etc. That means for each loss computation,
    we need to pull out the attribute of interest from each structure.
    E.g. oc20_energy from OC20 structures, oc22_energy from OC22
    structures etc. Set up those mappings here. Supports two kinds for
    now: 1) for each structure-level output, mapping from output head
    to boolean indexing map for `out` and `batch`, s.t. we can index like
    batch.oc20_energy[oc20_map] for oc20_energy loss calculation. 2) for
    each atom-level output, a similar mapping from output head to boolean
    indexing map. s.t. we can index like batch.oc20_forces[oc20_map].
    """

    output_masks = {task.name: torch.isfinite(batch[task.name])}
    if "forces" in task.name:
        output_masks[task.name] = output_masks[task.name].all(dim=1)

    for dset in set(batch.dataset_name):
        dset_mask = torch.from_numpy(np.array(batch.dataset_name) == dset).to(
            batch.pos.device
        )
        if task.level == "atom":
            dset_mask = torch.repeat_interleave(dset_mask, batch.natoms)
            output_masks[f"{dset}.{task.name}"] = dset_mask & output_masks[task.name]
        elif "stress" in task.name:
            assert output_masks[task.name].shape[0] == dset_mask.shape[0]
            # we need to expand the target mask shape to match the output shape
            target_shape = output_masks[task.name].shape
            dset_expanded = dset_mask.view(len(dset_mask), -1).expand(target_shape)
            output_masks[f"{dset}.{task.name}"] = (
                dset_expanded & output_masks[task.name]
            )
        else:
            output_masks[f"{dset}.{task.name}"] = dset_mask & output_masks[task.name]

    return output_masks


def get_output_masks(
    batch: AtomicData, tasks: Sequence[Task]
) -> dict[str, torch.Tensor]:
    """Same as above but for a list of tasks."""
    output_masks = {}
    for task in tasks:
        output_masks.update(get_output_mask(batch, task))

    return output_masks


def compute_loss(
    tasks: Sequence[Task], predictions: dict[str, torch.Tensor], batch: AtomicData
) -> dict[str, float]:
    """Compute loss given a sequence of tasks

    Args:
        tasks: a sequence of Task
        predictions: dictionary of predictions
        batch: data batch

    Returns:
        dictionary of losses for each task
    """

    batch_size = batch.natoms.numel()
    num_atoms_in_batch = batch.natoms.sum()

    free_mask = batch.fixed == 0
    output_masks = get_output_masks(batch, tasks)

    loss_dict = {}
    for task in tasks:
        # TODO this might be a very expensive clone
        target = batch[task.name].clone()
        output_mask = output_masks[task.name]

        # element references are applied to the target before normalization
        # TODO the current implementation will not work for single tasks with
        # multiple element references or normalizers
        # apply element references to the target
        if task.element_references is not None:
            with record_function("element_refs"):
                target = task.element_references.apply_refs(batch, target)
        # Normalize the target
        target = task.normalizer.norm(target)

        # Setting up a mult_mask to multiply the loss by 1 for valid atoms
        # or structures, and by 0 for the others. This is better than
        # indexing the loss with the mask, because it ensures that the
        # computational graph is correct.

        # this is related to how Hydra outputs stuff in nested dicts:
        # ie: oc20_energy.energy
        pred_for_task = predictions[task.name][task.property]
        if task.level == "atom":
            pred_for_task = pred_for_task.view(num_atoms_in_batch, -1)
        else:
            pred_for_task = pred_for_task.view(batch_size, -1)

        if task.level == "atom" and task.train_on_free_atoms:
            mult_mask = free_mask & output_mask
        else:
            mult_mask = output_mask
        loss_dict[task.name] = task.loss_fn(
            pred_for_task,
            target,
            mult_mask=mult_mask,
            natoms=batch.natoms,
        )

    # Sanity check to make sure the compute graph is correct.
    for lc in loss_dict.values():
        assert hasattr(lc, "grad_fn")

    return loss_dict


def compute_metrics(
    task: Task,
    predictions: dict[str, torch.Tensor],
    batch: AtomicData,
    dataset_name: str | None = None,
) -> dict[str:Metrics]:
    """Compute metrics and update running metrics for a given task

    Args:
        task: a Task
        predictions: dictionary of predictions
        batch: data batch
        dataset_name: optional, if given compute metrics for given task using only labels from the given dataset
        running_metrics: optional dictionary of previous metrics to update.

    Returns:
        dictionary of (updated) metrics
    """
    # output masks include task level mask, and task.dataset level masks.
    mask_key = task.name if dataset_name is None else f"{dataset_name}.{task.name}"
    output_mask = get_output_mask(batch, task)[mask_key]

    natoms = torch.repeat_interleave(batch.natoms, batch.natoms)
    if task.level == "atom":
        if task.eval_on_free_atoms is True:
            output_mask = output_mask & (batch.fixed == 0)

        natoms_masked = natoms[output_mask]
        output_size = natoms_masked.numel()
    elif "stress" in task.name:
        natoms_masked = batch.natoms[output_mask.all(dim=1)]
        output_size = output_mask.sum()
    else:
        natoms_masked = batch.natoms[output_mask]
        output_size = output_mask.sum()

    # no metrics to report
    if output_size == 0:
        return {metric_name: Metrics() for metric_name in task.metrics}

    target_masked = batch[task.name][output_mask]
    pred = predictions[task.name][task.property].clone()
    # denormalize the prediction
    pred = task.normalizer.denorm(pred)
    # undo element references for energy tasks
    if task.element_references is not None:
        pred = task.element_references.undo_refs(
            batch,
            pred,
        )
    pred_masked = pred[output_mask]

    # reshape: (num_atoms_in_batch, -1) or (num_systems_in_batch, -1)
    # if task.level == "atom" or "stress" not in task.name:
    #     # TODO do not reshape based on task.name
    #     # tensor.view(..., -1) will add an extra dimension even if the input shape is the same as output shape
    #     # this will cause downstream broadcast operations to be wrong and is dangerous
    #     target_masked = target_masked.view(output_size, -1)

    assert (
        target_masked.shape == pred_masked.shape
    ), f"shape mismatch for {task} target: target: {target_masked.shape}, pred: {pred_masked.shape}"

    # TODO need a cleaner interface for this...
    # Lets package up the masked target and prediction into a dictionary,
    # so that it plays nicely with the metrics functions
    # this does not work for metrics that use more than a single prediction key, ie energy_forces_within_threshold
    target_dict = {task.property: target_masked, "natoms": natoms_masked}
    pred_dict = {task.property: pred_masked}

    # this is different from original mt trainer, it assumes a Task has a single normalizer\
    # (even if it is used across datasets)
    metrics = {}
    for metric_name in task.metrics:
        # now predict the metrics and update them
        metric_fn = get_metrics_fn(metric_name)
        metrics[metric_name] = metric_fn(
            pred_dict, target_dict, key=task.property
        )  # TODO change this to return Metrics dataclass

    return metrics


def mt_collater_adapter(
    tasks: list[Task], exclude_keys: list[str] = DEFAULT_EXCLUDE_KEYS
):
    # this is required because the MTCollater needs the old json formated task config so we need to convert it here
    task_config_old = {}
    tasks = filter_inference_only_tasks(tasks)
    for task in tasks:
        task_config_old[task.name] = {
            "level": task.level,
            "property": task.property,
            "out_spec": {
                "dim": list(task.out_spec.dim),
                "dtype": str(task.out_spec.dtype),
            },
            "datasets": task.datasets,
            "train_on_free_atoms": task.train_on_free_atoms,
            "eval_on_free_atoms": task.eval_on_free_atoms,
        }
    return MTCollater(task_config_old, exclude_keys)


def _get_consine_lr_scheduler(
    warmup_factor: float,
    warmup_epochs: float,
    lr_min_factor: float,
    n_iters_per_epoch: int,
    optimizer: torch.optim.Optimizer,
    epochs: Optional[int] = None,
    steps: Optional[int] = None,
) -> torch.optim.lr_scheduler.LRScheduler:
    assert (epochs is not None) ^ (
        steps is not None
    ), "Exactly one of epochs or steps must be None/Not-None (XOR)"
    scheduler_steps = int(epochs * n_iters_per_epoch) if steps is None else steps
    # fixed function for constructing a LambdaLR scheduler
    lambda_fn = CosineLRLambda(
        warmup_epochs=max(
            int(warmup_epochs * n_iters_per_epoch), 1
        ),  # this cannot be 0
        warmup_factor=warmup_factor,
        epochs=scheduler_steps,
        lr_min_factor=lr_min_factor,
    )
    return torch.optim.lr_scheduler.LambdaLR(optimizer, lambda_fn)


def _get_optimizer_wd(
    optimizer_fn: callable, model: torch.nn.Module
) -> torch.optim.Optimizer:
    weight_decay = optimizer_fn.keywords.get("weight_decay", 0)
    # split the params into the params with and without WD
    # some fairchem models implement a no_weight_decay and this
    # is used to return params such as embeddings that should have no wd.
    # TODO: use a protocol here instead of guessing that the attr exists
    if weight_decay > 0 and hasattr(model, "no_weight_decay"):
        model_params_no_wd = model.no_weight_decay()

        params_decay, params_no_decay, name_no_decay = [], [], []
        for name, param in model.named_parameters():
            if not param.requires_grad:
                continue

            if any(name.endswith(skip_name) for skip_name in model_params_no_wd):
                params_no_decay.append(param)
                name_no_decay.append(name)
            else:
                params_decay.append(param)

        if distutils.is_master():
            logging.info("Parameters without weight decay:")
            logging.info(name_no_decay)

        optimizer = optimizer_fn(
            params=[
                {"params": params_no_decay, "weight_decay": 0},
                {"params": params_decay, "weight_decay": weight_decay},
            ]
        )
    else:
        optimizer = optimizer_fn(params=model.parameters())
    return optimizer


def _reshard_fsdp(model: torch.nn.Module) -> None:
    for m in FullyShardedDataParallel.fsdp_modules(model):
        if m._has_params and m.sharding_strategy is not ShardingStrategy.NO_SHARD:
            torch.distributed.fsdp._runtime_utils._reshard(m, m._handle, True)


def set_sampler_state(state: State, epoch: int, step_start: int) -> None:
    logging.info(
        f"at beginning of epoch {epoch}, setting sampler start step to {step_start}"
    )
    if hasattr(state.train_state.dataloader, "batch_sampler"):
        # the batch sampler must have the set_epoch_and_start_iteration to be able to restore state
        assert hasattr(
            state.train_state.dataloader.batch_sampler,
            "set_epoch_and_start_iteration",
        )
        state.train_state.dataloader.batch_sampler.set_epoch_and_start_iteration(
            epoch, step_start
        )
    else:
        logging.warning(
            "AtomicData sampler not found in dataloader, no dataloader state restored!!"
        )


class MLIPTrainEvalUnit(
    TrainUnit[AtomicData], EvalUnit[AtomicData], Stateful, Checkpointable
):
    def __init__(
        self,
        job_config: DictConfig,
        model: torch.nn.Module,
        optimizer_fn: callable,
        cosine_lr_scheduler_fn: callable,
        tasks: list[Task],
        bf16: bool = False,
        print_every: int = 10,
        clip_grad_norm: float | None = None,
        ema_decay: float = 0.999,
        train_strategy: TrainStrategy = TrainStrategy.DDP,
        debug_checksums_save_path: str | None = None,
        profile_flops: bool = False,
        save_inference_ckpt: bool = True,
    ):
        super().__init__()
        self.job_config = job_config
        # throw out tasks that are inference_only (don't use them for training/eval)
        self.tasks = filter_inference_only_tasks(tasks)
        self.profile_flops = profile_flops
        self.save_inference_ckpt = save_inference_ckpt

        for task in self.tasks:
            if task.element_references is not None:
                task.element_references.to(torch.device(get_device_for_local_rank()))

        # placeholder for autocast code, may need to move out to common
        self.bf16 = bf16
        self.autocast_enabled = self.bf16
        self.autocast_dtype = torch.bfloat16

        self.finetune_model_full_config = getattr(
            model, "finetune_model_full_config", None
        )

        # call optimizer function between wrapping in DDP
        # this is required for models that have a no_weight_decay function
        self.optimizer = _get_optimizer_wd(optimizer_fn, model)

        self.logger = (
            WandBSingletonLogger.get_instance()
            if distutils.is_master()
            and not self.job_config.debug
            and self.job_config.logger
            else None
        )
        self.debug_checksums_save_path = debug_checksums_save_path
        if self.debug_checksums_save_path:
            os.makedirs(debug_checksums_save_path, exist_ok=True)
        self.print_every = print_every
        self.clip_grad_norm = clip_grad_norm
        self.dp_world_size = (
            gp_utils.get_dp_world_size()
            if gp_utils.initialized()
            else distutils.get_world_size()
        )

        self.num_params = sum(p.numel() for p in model.parameters())
        if self.logger:
            self.logger.log_summary(
                {"num_params": self.num_params, "dp_world_size": self.dp_world_size}
            )

        model.to(torch.device(get_device_for_local_rank()))

        self.ema_decay = ema_decay
        self.ema_model = None
        self.train_strategy = train_strategy
        if train_strategy == TrainStrategy.DDP:
            self.model = prepare_module(
                model, device=torch.device(get_device_for_local_rank()), strategy="ddp"
            )
            if self.ema_decay is not None:
                self.ema_model = torch.optim.swa_utils.AveragedModel(
                    self.model,
                    multi_avg_fn=torch.optim.swa_utils.get_ema_multi_avg_fn(
                        self.ema_decay
                    ),
                )
        elif train_strategy == TrainStrategy.FSDP:
            # only wrap MOElinears for now, these are the layers that has large parameter size
            from fairchem.core.models.uma.escn_moe import MOELinear

            shard_group_size = job_config.scheduler.ranks_per_node
            mesh_2d = init_device_mesh(
                get_device_for_local_rank(),
                mesh_shape=(
                    int(distutils.get_world_size() // shard_group_size),
                    shard_group_size,
                ),
                mesh_dim_names=("replicate", "shard"),
            )
            fsdp_params = {
                "sharding_strategy": ShardingStrategy.HYBRID_SHARD,
                "device_mesh": mesh_2d,
                "auto_wrap_policy": ModuleWrapPolicy(module_classes=[MOELinear]),
                "use_orig_params": True,
            }
            if self.ema_decay is not None:
                ema_model = torch.optim.swa_utils.AveragedModel(
                    model,
                    multi_avg_fn=torch.optim.swa_utils.get_ema_multi_avg_fn(
                        self.ema_decay
                    ),
                )
                self.ema_model = FullyShardedDataParallel(ema_model, **fsdp_params)

                FullyShardedDataParallel.set_state_dict_type(
                    self.ema_model,
                    StateDictType.SHARDED_STATE_DICT,
                    state_dict_config=ShardedStateDictConfig(),
                )
            self.model = FullyShardedDataParallel(model, **fsdp_params)
            FullyShardedDataParallel.set_state_dict_type(
                self.model,
                StateDictType.SHARDED_STATE_DICT,
                state_dict_config=ShardedStateDictConfig(),
            )

            logging.info(f"Create device mesh {mesh_2d} for FSDP")
        else:
            raise ValueError(f"Unknown Training Strategy {train_strategy}")

        # setup eval unit, make it share the same model as this unit and turn off the logger
        self.eval_unit = MLIPEvalUnit(job_config=job_config, model=None, tasks=tasks)
        eval_model = self.ema_model if self.ema_model is not None else self.model
        self.eval_unit.setup_train_eval_unit(eval_model)

        self.cosine_lr_scheduler_fn = cosine_lr_scheduler_fn
        self.scheduler = None
        self.lazy_state_location = None

    def load_scheduler(self, train_dataloader_size: int) -> int:
        self.scheduler = self.cosine_lr_scheduler_fn(
            n_iters_per_epoch=train_dataloader_size,
            optimizer=self.optimizer,
        )

    def on_train_start(self, state: State) -> None:
        self.model.train()
        if self.profile_flops:
            # runtime import to make this feature optional
            from fairchem.core.components.common.flops_profile import get_flops_profile

            data = next(iter(state.train_state.dataloader)).to(
                get_device_for_local_rank()
            )
            flops = get_flops_profile(self.model, data, verbose=True)
            num_atoms_local = data.natoms.sum().item()
            flops_per_atom_param = flops / self.num_params / num_atoms_local
            if self.logger:
                self.logger.log_summary(
                    {
                        "train/fwd_flops": flops,
                        "train/fwd_flops_per_atom_param": flops_per_atom_param,
                    }
                )
            if "edge_index" in data:
                num_edges_local = data.edge_index.shape[1]
                flops_per_edge_param = flops / self.num_params / num_edges_local
                if self.logger:
                    self.logger.log_summary(
                        {
                            "train/fwd_flops_per_edge_param": flops_per_edge_param,
                            "train/num_edges_local": num_edges_local,
                        }
                    )

        if self.scheduler is None:
            self.load_scheduler(len(state.train_state.dataloader))

        if self.lazy_state_location is not None:
            self._execute_load_state(self.lazy_state_location)

        self.previous_wall_time = time.time()
        # this should only be non-zero if we are resuming from a run
        epoch = self.train_progress.num_epochs_completed
        start_step = self.train_progress.num_steps_completed_in_epoch
        logging.info(f"on_train_start: setting sampler state to {epoch}, {start_step}")
        set_sampler_state(
            state,
            epoch,
            start_step,
        )

    def on_train_epoch_start(self, state: State) -> None:
        # we can safely set start steps to 0 here because this callback is NOT called when resuming from a run
        # https://github.com/pytorch/tnt/blob/master/torchtnt/framework/train.py#L187
        set_sampler_state(state, self.train_progress.num_epochs_completed, 0)

    def train_step(self, state: State, data: AtomicData) -> None:
        try:
            device = get_device_for_local_rank()
            batch_on_device = data.to(device)
            step = self.train_progress.num_steps_completed
            epoch = (
                self.train_progress.num_epochs_completed
                + self.train_progress.num_steps_completed_in_epoch
                / float(len(state.train_state.dataloader))
            )
            with torch.autocast(
                device_type=device,
                enabled=self.autocast_enabled,
                dtype=self.autocast_dtype,
            ):
                with record_function("forward"):
                    pred = self.model.forward(batch_on_device)
                with record_function("compute_loss"):
                    loss_dict = compute_loss(self.tasks, pred, batch_on_device)
            scalar_loss = sum(loss_dict.values())
            self.optimizer.zero_grad()
            with record_function("backward"):
                scalar_loss.backward()

            if self.debug_checksums_save_path:
                gp_size = 0
                gp_rank = 0
                if gp_utils.initialized():
                    gp_size = gp_utils.get_dp_world_size()
                    gp_rank = gp_utils.get_dp_rank()

                ddp_size = distutils.get_world_size()
                ddp_rank = distutils.get_rank()

                fn = os.path.join(
                    self.debug_checksums_save_path,
                    f"ddp{ddp_size}.{ddp_rank}_gp{gp_size}.{gp_rank}_step{step}.txt",
                )
                with open(fn, "w") as f:
                    f.write(f"Loss,{scalar_loss.item()}\n")
                    for name, param in self.model.named_parameters():
                        if param.grad is not None:
                            f.write(
                                f"Param,{step},{name},{param.abs().mean().item()}\n"
                            )
                            f.write(
                                f"Grad,{step},{name},{param.grad.abs().mean().item()}\n"
                            )

            if self.clip_grad_norm is not None:
                if self.train_strategy == TrainStrategy.FSDP:
                    grad_norm = self.model.clip_grad_norm_(
                        max_norm=self.clip_grad_norm,
                    )
                else:
                    grad_norm = torch.nn.utils.clip_grad_norm_(
                        self.model.parameters(),
                        max_norm=self.clip_grad_norm,
                    )

                if self.logger:
                    self.logger.log({"train/grad_norm": grad_norm}, step=step)
            self.optimizer.step()
            if self.ema_model is not None:
                self.ema_model.update_parameters(self.model)

            self.optimizer.zero_grad(set_to_none=True)

            time_delta = time.time() - self.previous_wall_time
            self.previous_wall_time = time.time()
            num_atoms_local = data.natoms.sum().item()
            num_samples_local = data.natoms.numel()
            log_dict = {
                "train/loss": scalar_loss.item(),
                "train/lr": self.scheduler.get_lr()[0],
                "train/step": step,
                "train/epoch": epoch,
                "train/samples_per_second(approx)": num_samples_local
                * self.dp_world_size
                / float(time_delta),
                "train/atoms_per_second(approx)": num_atoms_local
                * self.dp_world_size
                / float(time_delta),
                "train/num_atoms_on_rank": num_atoms_local,
                "train/num_samples_on_rank": num_samples_local,
            }

            if self.logger:
                self.logger.log(log_dict, step=step, commit=True)

            if step % self.print_every == 0:
                logging.info(log_dict)

            self.scheduler.step()

            # TODO: compute metrics
            self.last_loss = scalar_loss.item()
        except Exception:
            logging.error(
                f"Exception during training! On step {self.train_progress.num_steps_completed}"
            )
            logging.error(
                f"Data info: {data}\ndata.dataset: {data.dataset}\n"
                + f"data.sid: {data.sid if 'sid' in data else None}\n"
                + f"data.natoms: {data.natoms}\n"
                + f"data.atomic_numbers: {data.atomic_numbers}"
            )
            raise

    def on_train_end(self, state: State) -> None:
        logging.info(
            f"Training Completed {self.train_progress.num_steps_completed} steps"
        )

    def state_dict(
        self,
    ) -> dict[str, Any]:
        # this line automatically manages FSDP FQN's, as well as sets the default state dict type to FSDP.SHARDED_STATE_DICT
        model_state_dict, optimizer_state_dict = get_state_dict(
            self.model, self.optimizer
        )
        ema_state_dict = (
            get_model_state_dict(self.ema_model) if self.ema_model else None
        )
        state = {
            "progress": self.train_progress.state_dict(),
            "model": model_state_dict,
            "ema": ema_state_dict,
            "optim": optimizer_state_dict,
            "scheduler": self.scheduler.state_dict(),
        }
        return state

    def load_state_dict(self, state_dict: dict[str, Any]):
        # sets our state dicts on the model and optimizer, now that we've loaded
        set_state_dict(
            self.model,
            self.optimizer,
            model_state_dict=state_dict["model"],
            optim_state_dict=state_dict["optim"],
        )
        self.train_progress.load_state_dict(state_dict["progress"])
        self.scheduler.load_state_dict(state_dict["scheduler"])
        if self.ema_model is not None:
            set_model_state_dict(
                self.ema_model,
                model_state_dict=state_dict["ema"],
            )
            self.ema_model.load_state_dict(state_dict["ema"])

    def eval_step(self, state: State, data: AtomicData) -> None:
        self.eval_unit.eval_step(state, data)

    def on_eval_epoch_start(self, state: State) -> None:
        self.eval_unit.on_eval_epoch_start(state)

    def on_eval_epoch_end(self, state: State) -> None:
        metrics = self.eval_unit.on_eval_epoch_end(state)
        if self.logger is not None:
            self.logger.log(metrics, commit=False)
        # Need to manually reshard the FSDP ema model: https://github.com/pytorch/pytorch/issues/117421#issuecomment-1890948734, otherwise we don't update the ema model weights correctly
        if self.ema_model and self.train_strategy == TrainStrategy.FSDP:
            _reshard_fsdp(self.ema_model)

    def get_finetune_model_config(self) -> DictConfig | None:
        return self.finetune_model_full_config

    def save_state(self, checkpoint_location: str) -> None:
        # save a resume config that can be easily used for resuming runs
        os.makedirs(checkpoint_location, exist_ok=True)
        config = OmegaConf.load(self.job_config.metadata.config_path)
        config.job.runner_state_path = checkpoint_location

        finetune_model_full_config = self.get_finetune_model_config()
        if finetune_model_full_config is not None:
            config.runner.train_eval_unit.model = finetune_model_full_config

        OmegaConf.save(config, os.path.join(checkpoint_location, UNIT_RESUME_CONFIG))

        # calls train_eval_unit.save_state
        state = {"unit_state": self.state_dict(), "config": config}
        dcp.save(state, checkpoint_id=checkpoint_location)

        # warning this can be VERY SLOW for large models, better not to do this at every checkpoint
        if (
            self.save_inference_ckpt
            and distutils.is_master()
            and os.path.exists(checkpoint_location)
        ):
            convert_train_checkpoint_to_inference_checkpoint(
                checkpoint_location,
                os.path.join(checkpoint_location, UNIT_INFERENCE_CHECKPOINT),
            )

        logging.info(f"Saved dcp checkpoint to {checkpoint_location}")

    def load_state(self, checkpoint_location: str | None) -> None:
        # lazily load state, save the state and load on the first train step
        self.lazy_state_location = checkpoint_location

    def _execute_load_state(self, checkpoint_location: str | None) -> None:
        state = {"unit_state": self.state_dict()}
        dcp.load(state_dict=state, checkpoint_id=checkpoint_location)
        self.load_state_dict(state["unit_state"])
        logging.info(f"Done loading checkpoint from {checkpoint_location}")


class MLIPEvalUnit(EvalUnit[AtomicData]):
    def __init__(
        self,
        job_config: DictConfig,
        model: torch.nn.Module,
        tasks: Sequence[Task],
        bf16: bool = False,
    ):
        """Evaluate your MLIPs and so forth.

        Args:
            job_config: a job config object specifying logger and job type
            model: model to evaluate
            evaluations: a list of evaluation objects
            bf16: whether to use autocast with bf16
        """
        super().__init__()
        self.job_config = job_config
        self.model = model
        self.tasks = filter_inference_only_tasks(tasks)

        for task in self.tasks:
            if task.element_references is not None:
                task.element_references.to(torch.device(get_device_for_local_rank()))

        # dictionary of metrics for each dataset, split, task, and metric
        self.running_metrics: dict[str, dict[str, dict[str, Metrics]]] = {}
        self.total_loss_metrics: Metrics = Metrics()
        self.total_atoms: int = 0
        self.total_runtime: float = 0

        # allow the model to be set separately (this is used by the TrainEvalunit to initialize one model for both train and eval)
        if self.model is not None:
            self.model = prepare_module(
                model, device=torch.device(get_device_for_local_rank()), strategy="ddp"
            )
        self.logger = (
            WandBSingletonLogger.get_instance()
            if distutils.is_master()
            and not self.job_config.debug
            and self.job_config.logger
            else None
        )

        # TODO see placeholder comment in TrainEvalUnit as well
        self.autocast_enabled = bf16
        self.autocast_dtype = torch.bfloat16

    def setup_train_eval_unit(self, model: torch.nn.Module) -> None:
        self.model = model
        self.logger = None

    def on_eval_epoch_start(self, state: State) -> None:
        """Reset all metrics, and make sure model is in eval mode."""
        # TODO store ema here as well?
        self.model.eval()

        # create dictionary of running metrics with following schema:
        # task.name: {dataset.split: {metric: value}}}
        datasets_to_eval = state.eval_state.dataloader.dataset.dataset_names
        self.running_metrics = {
            task.name: {
                dataset: {metric: Metrics() for metric in task.metrics}
                for dataset in filter(
                    lambda x: any(dset in x for dset in task.datasets), datasets_to_eval
                )
            }
            for task in self.tasks
        }
        self.total_loss_metrics = Metrics()
        self.total_atoms = 0
        self.total_runtime = 0
        self.total_len = len(state.eval_state.dataloader)
        self.start_time = time.time()
        self.last_report = time.time()
        self.report_every = 180

    def eval_step(self, state: State, data: AtomicData) -> None:
        """Evaluates the model on a batch of data."""
        device = get_device_for_local_rank()
        data = data.to(device)
        self.total_atoms += data.natoms.sum().item()

        if (time.time() - self.last_report) > self.report_every:
            seconds_per_step = (time.time() - self.start_time) / max(
                1, self.eval_progress.num_steps_completed
            )
            eta_hours = self.total_len * seconds_per_step / (60.0 * 60.0)
            print(
                f"step: {self.eval_progress.num_steps_completed}, seconds_per_step: {seconds_per_step} eta_hours: {eta_hours}"
            )
            self.last_report = time.time()

        with torch.autocast(
            device_type=get_device_for_local_rank(),
            enabled=self.autocast_enabled,
            dtype=self.autocast_dtype,
        ):
            t0 = time.time()
            preds = self.model(data)
            self.total_runtime += time.time() - t0

        # compute the loss
        loss_dict = compute_loss(self.tasks, preds, data)
        total_loss = sum(loss_dict.values())
        self.total_loss_metrics += Metrics(metric=total_loss, total=total_loss, numel=1)

        # get the datasets with split names
        datasets_in_batch = set(data.dataset_name)

        # run each evaluation
        for task in self.tasks:
            # This filters out all the datasets.splits that are in the current batch and
            # and are also included in the task. Remove the dataset.splits not in the task, since we
            # wont compute metrics for those.
            # TODO overhaul the dataset names in task to avoid this filter?
            for dataset in filter(
                lambda x: any(dset_name in x for dset_name in task.datasets),
                datasets_in_batch,
            ):
                current_metrics = compute_metrics(task, preds, data, dataset)
                running_metrics = self.running_metrics[task.name][dataset]

                for metric_name in task.metrics:
                    running_metrics[metric_name] += current_metrics[metric_name]

                self.running_metrics[task.name][dataset].update(running_metrics)

                # update the loss metrics
                # loss_metrics = Metrics(
                #     metric=loss_dict[task.name], total=loss_dict[task.name], numel=1
                # )
                # self.running_metrics[task.name][dataset]["loss"] += loss_metrics

                # # total loss
                # loss = sum(loss_dict.values()).item()
                # self.total_loss_metrics += Metrics(metric=loss, total=loss, numel=1)

    def on_eval_epoch_end(self, state: State) -> dict:
        """Aggregate all metrics and log."""

        logging.info("Done eval epoch, aggregating metrics")
        device = get_device_for_local_rank()
        log_dict = {}
        for task, dataset_dict in self.running_metrics.items():
            for dataset, metrics_dict in dataset_dict.items():
                for metric_name, metrics in metrics_dict.items():
                    total = distutils.all_reduce(
                        metrics.total, average=False, device=device
                    )
                    numel = distutils.all_reduce(
                        metrics.numel, average=False, device=device
                    )
                    log_dict[f"val/{dataset},{task},{metric_name}"] = total / numel

        total_runtime = distutils.all_reduce(
            self.total_runtime, average=False, device=device
        )
        total_atoms = distutils.all_reduce(
            self.total_atoms, average=False, device=device
        )

        # we do not reduce across ranks here. DDP loss uses a _ddp_mean that
        # gives the an approximate loss -> loss_rank_i / average_num_samples_across_ranks
        log_dict["val/loss"] = self.total_loss_metrics.metric
        log_dict["val/atoms_per_second"] = total_atoms / total_runtime
        log_dict["val/epoch"] = self.eval_progress.num_epochs_completed

        if self.logger is not None:
            self.logger.log(log_dict, commit=True)

        log_str = "".join(
            f"  {k}: {log_dict[k]:.4f}\n" for k in sorted(log_dict.keys())
        )
        logging.info(f"Finished aggregating metrics: \n{log_str}")

        return log_dict



from __future__ import annotations

import dataclasses

import torch


@dataclasses.dataclass
class GraphAttentionData:
    """
    Custom dataclass for storing graph data for Graph Attention Networks
    atomic_numbers: (N)
    edge_distance_expansion: (N, max_nei, edge_distance_expansion_size)
    edge_direction: (N, max_nei, 3)
    node_direction_expansion: (N, node_direction_expansion_size)
    attn_mask: (N * num_head, max_nei, max_nei) Attention mask with angle embeddings
    angle_embedding: (N * num_head, max_nei, max_nei) Angle embeddings (cosine)
    frequency_vectors: (N, max_nei, head_dim, 2l+1) Frequency embeddings
    neighbor_list: (N, max_nei)
    neighbor_mask: (N, max_nei)
    node_batch: (N)
    node_padding_mask: (N)
    graph_padding_mask: (num_graphs)
    """

    atomic_numbers: torch.Tensor
    edge_distance_expansion: torch.Tensor
    edge_direction: torch.Tensor
    node_direction_expansion: torch.Tensor
    attn_mask: torch.Tensor
    angle_embedding: torch.Tensor | None
    frequency_vectors: torch.Tensor | None
    neighbor_list: torch.Tensor
    neighbor_mask: torch.Tensor
    node_batch: torch.Tensor
    node_padding_mask: torch.Tensor
    graph_padding_mask: torch.Tensor


def map_graph_attention_data_to_device(
    data: GraphAttentionData, device: torch.device | str
) -> GraphAttentionData:
    """
    Map all tensor fields in GraphAttentionData to the specified device.
    """
    kwargs = {}
    for field in dataclasses.fields(data):
        field_value = getattr(data, field.name)
        if isinstance(field_value, torch.Tensor):
            kwargs[field.name] = field_value.to(device)
        elif field_value is None:
            kwargs[field.name] = None
        else:
            # Handle any other types that might be added in the future
            kwargs[field.name] = field_value

    return GraphAttentionData(**kwargs)


def flatten_graph_attention_data_with_spec(data, spec):
    # Flatten based on the in_spec structure
    flat_data = []
    for field_name in spec.context[0]:
        field_value = getattr(data, field_name)
        if isinstance(field_value, torch.Tensor):
            flat_data.append(field_value)
        elif field_value is None:
            flat_data.append(None)
        else:
            # Handle custom types like AttentionBias
            flat_data.extend(field_value.tree_flatten())
    return tuple(flat_data)


torch.export.register_dataclass(
    GraphAttentionData, serialized_type_name="GraphAttentionData"
)
torch.fx._pytree.register_pytree_flatten_spec(  # type: ignore
    GraphAttentionData, flatten_fn_spec=flatten_graph_attention_data_with_spec
)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from typing import TypeVar

_T = TypeVar("_T")


def assert_is_instance(obj: object, cls: type[_T]) -> _T:
    if obj and not isinstance(obj, cls):
        raise TypeError(f"obj is not an instance of cls: obj={obj}, cls={cls}")
    return obj


def none_throws(x: _T | None, msg: str | None = None) -> _T:
    if x is None:
        if msg:
            raise ValueError(msg)
        raise ValueError("x cannot be None")
    return x



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import math

import numpy as np
import numpy.typing as npt
import torch
import torch.nn as nn
from scipy.special import sph_harm_y

from .activations import Act


class Sine(nn.Module):
    def __init__(self, w0: float = 30.0) -> None:
        super().__init__()
        self.w0 = w0

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.sin(self.w0 * x)


class SIREN(nn.Module):
    def __init__(
        self,
        layers: list[int],
        num_in_features: int,
        out_features: int,
        w0: float = 30.0,
        initializer: str | None = "siren",
        c: float = 6,
    ) -> None:
        super().__init__()
        self.layers = [nn.Linear(num_in_features, layers[0]), Sine(w0=w0)]

        for index in range(len(layers) - 1):
            self.layers.extend(
                [nn.Linear(layers[index], layers[index + 1]), Sine(w0=1)]
            )

        self.layers.append(nn.Linear(layers[-1], out_features))
        self.network = nn.Sequential(*self.layers)

        if initializer is not None and initializer == "siren":
            for m in self.network:
                if isinstance(m, nn.Linear):
                    num_input = float(m.weight.size(-1))
                    with torch.no_grad():
                        m.weight.uniform_(
                            -math.sqrt(6.0 / num_input),
                            math.sqrt(6.0 / num_input),
                        )

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        return self.network(X)


class SINESmearing(nn.Module):
    def __init__(
        self,
        num_in_features: int,
        num_freqs: int = 40,
        use_cosine: bool = False,
    ) -> None:
        super().__init__()

        self.num_freqs = num_freqs
        self.out_dim: int = num_in_features * self.num_freqs
        self.use_cosine = use_cosine

        freq = torch.arange(num_freqs).float()
        freq = torch.pow(torch.ones_like(freq) * 1.1, freq)
        self.freq_filter = nn.Parameter(
            freq.view(-1, 1).repeat(1, num_in_features).view(1, -1),
            requires_grad=False,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.repeat(1, self.num_freqs)
        x = x * self.freq_filter

        if self.use_cosine:
            return torch.cos(x)
        else:
            return torch.sin(x)


class GaussianSmearing(nn.Module):
    def __init__(
        self,
        num_in_features: int,
        start: int = 0,
        end: int = 1,
        num_freqs: int = 50,
    ) -> None:
        super().__init__()
        self.num_freqs = num_freqs
        offset = torch.linspace(start, end, num_freqs)
        self.coeff: float = -0.5 / (offset[1] - offset[0]).item() ** 2
        self.offset = nn.Parameter(
            offset.view(-1, 1).repeat(1, num_in_features).view(1, -1),
            requires_grad=False,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.repeat(1, self.num_freqs)
        x = x - self.offset
        return torch.exp(self.coeff * torch.pow(x, 2))


class FourierSmearing(nn.Module):
    def __init__(
        self,
        num_in_features: int,
        num_freqs: int = 40,
        use_cosine: bool = False,
    ) -> None:
        super().__init__()

        self.num_freqs = num_freqs
        self.out_dim: int = num_in_features * self.num_freqs
        self.use_cosine = use_cosine

        freq = torch.arange(num_freqs).to(torch.float32)
        self.freq_filter = nn.Parameter(
            freq.view(-1, 1).repeat(1, num_in_features).view(1, -1),
            requires_grad=False,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.repeat(1, self.num_freqs)
        x = x * self.freq_filter

        if self.use_cosine:
            return torch.cos(x)
        else:
            return torch.sin(x)


class Basis(nn.Module):
    smearing: SINESmearing | FourierSmearing | GaussianSmearing | torch.nn.Sequential

    def __init__(
        self,
        num_in_features: int,
        num_freqs: int = 50,
        basis_type: str = "powersine",
        act: str = "ssp",
        sph: SphericalSmearing | None = None,
    ) -> None:
        super().__init__()

        self.num_freqs = num_freqs
        self.basis_type = basis_type

        if basis_type == "powersine":
            self.smearing = SINESmearing(num_in_features, num_freqs)
            self.out_dim = num_in_features * num_freqs
        elif basis_type == "powercosine":
            self.smearing = SINESmearing(num_in_features, num_freqs, use_cosine=True)
            self.out_dim = num_in_features * num_freqs
        elif basis_type == "fouriersine":
            self.smearing = FourierSmearing(num_in_features, num_freqs)
            self.out_dim = num_in_features * num_freqs
        elif basis_type == "gauss":
            self.smearing = GaussianSmearing(
                num_in_features, start=0, end=1, num_freqs=num_freqs
            )
            self.out_dim = num_in_features * num_freqs
        elif basis_type == "linact":
            self.smearing = torch.nn.Sequential(
                torch.nn.Linear(num_in_features, num_freqs * num_in_features),
                Act(act),
            )
            self.out_dim = num_in_features * num_freqs
        elif basis_type == "raw" or basis_type == "rawcat":
            self.out_dim = num_in_features
        elif "sph" in basis_type:
            # by default, we use sine function to encode distance
            # sph must be given here
            assert sph is not None
            # assumes the first three columns are normalizaed xyz
            # the rest of the columns are distances
            if "cat" in basis_type:
                # concatenate
                self.smearing_sine = SINESmearing(num_in_features - 3, num_freqs)
                self.out_dim = sph.out_dim + (num_in_features - 3) * num_freqs
            elif "mul" in basis_type:
                self.smearing_sine = SINESmearing(num_in_features - 3, num_freqs)
                self.lin = torch.nn.Linear(
                    self.smearing_sine.out_dim, num_in_features - 3
                )
                self.out_dim = (num_in_features - 3) * sph.out_dim
            elif "m40" in basis_type:
                dim = 40
                self.smearing_sine = SINESmearing(num_in_features - 3, num_freqs)
                self.lin = torch.nn.Linear(
                    self.smearing_sine.out_dim, dim
                )  # make the output dimensionality comparable.
                self.out_dim = dim * sph.out_dim
            elif "nosine" in basis_type:
                # does not use sine smearing for encoding distance
                self.out_dim = (num_in_features - 3) * sph.out_dim
            else:
                raise ValueError("cat or mul not specified for spherical harnomics.")
        else:
            raise RuntimeError("Undefined basis type.")

    def forward(self, x: torch.Tensor, edge_attr_sph: torch.Tensor | None = None):
        if "sph" in self.basis_type:
            if "nosine" not in self.basis_type:
                x_sine = self.smearing_sine(
                    x[:, 3:]
                )  # the first three features correspond to edge_vec_normalized, so we ignore
                if "cat" in self.basis_type:
                    # just concatenate spherical edge feature and sined node features
                    assert isinstance(edge_attr_sph, torch.Tensor)
                    return torch.cat([edge_attr_sph, x_sine], dim=1)
                elif "mul" in self.basis_type or "m40" in self.basis_type:
                    # multiply sined node features into spherical edge feature (inspired by theory in spherical harmonics)
                    r = self.lin(x_sine)
                    outer = torch.einsum("ik,ij->ikj", edge_attr_sph, r)
                    return torch.flatten(outer, start_dim=1)
                else:
                    raise RuntimeError(f"Unknown basis type called {self.basis_type}")
            else:
                outer = torch.einsum("ik,ij->ikj", edge_attr_sph, x[:, 3:])
                return torch.flatten(outer, start_dim=1)

        elif "raw" in self.basis_type:
            # do nothing, just return node features
            pass
        else:
            x = self.smearing(x)
        return x


class SphericalSmearing(nn.Module):
    m: npt.NDArray[np.int_]
    n: npt.NDArray[np.int_]

    def __init__(self, max_n: int = 10, option: str = "all") -> None:
        super().__init__()

        self.max_n = max_n

        m_list: list[int] = []
        n_list: list[int] = []
        for i in range(max_n):
            for j in range(i + 1):
                m_list.append(j)
                n_list.append(i)

        m = np.array(m_list)
        n = np.array(n_list)

        if option == "all":
            self.m = m
            self.n = n
        elif option == "sine":
            self.m = m[n % 2 == 1]
            self.n = n[n % 2 == 1]
        elif option == "cosine":
            self.m = m[n % 2 == 0]
            self.n = n[n % 2 == 0]

        self.out_dim = int(np.sum(self.m == 0) + 2 * np.sum(self.m != 0))

    def forward(self, xyz: torch.Tensor) -> torch.Tensor:
        # assuming input is already normalized
        assert xyz.size(1) == 3

        xyz = xyz / xyz.norm(dim=-1).view(-1, 1)

        phi = torch.acos(xyz[:, 2])
        theta = torch.atan2(-xyz[:, 1], -xyz[:, 0]) + math.pi

        phi = phi.cpu().numpy()
        theta = theta.cpu().numpy()

        m_tile = np.tile(self.m, (len(xyz), 1))
        n_tile = np.tile(self.n, (len(xyz), 1))
        theta_tile = np.tile(theta.reshape(len(xyz), 1), (1, len(self.m)))
        phi_tile = np.tile(phi.reshape(len(xyz), 1), (1, len(self.m)))

        harm = sph_harm_y(n_tile, m_tile, theta_tile, phi_tile)

        harm_mzero = harm[:, self.m == 0]
        harm_mnonzero = harm[:, self.m != 0]

        harm_real = np.concatenate(
            [harm_mzero.real, harm_mnonzero.real, harm_mnonzero.imag], axis=1
        )

        return torch.from_numpy(harm_real).to(torch.float32).to(xyz.device)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
from glob import glob
from typing import TYPE_CHECKING, Any

import pandas as pd
from ase.io.jsonio import decode
from monty.dev import requires
from tqdm import tqdm

from fairchem.core.components.benchmark import JsonDFReducer
from fairchem.core.components.calculate import (
    RelaxationRunner,
)
from fairchem.core.components.calculate.recipes.energy import calc_energy_from_e_refs

try:
    from matbench_discovery.enums import MbdKey
    from matbench_discovery.metrics.discovery import stable_metrics
    from matbench_discovery.metrics.geo_opt import calc_geo_opt_metrics
    from matbench_discovery.structure import symmetry
    from pymatgen.entries.compatibility import MaterialsProject2020Compatibility
    from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry
    from pymatgen.io.ase import AseAtomsAdaptor
    from pymatviz.enums import Key

    mbd_installed = True
    MP2020Compatibility = MaterialsProject2020Compatibility()

except ImportError:
    mbd_installed = False
    MP2020Compatibility = None

if TYPE_CHECKING:
    from pymatgen.core import Structure
    from pymatgen.entries.compatibility import Compatibility


def as_dict_handler(obj: Any) -> dict[str, Any] | None:
    """Pass this to json.dump(default=) or as pandas.to_json(default_handler=) to
    serialize Python classes with as_dict(). Warning: Objects without a as_dict() method
    are replaced with None in the serialized data.

    From matbench_discovery: https://github.com/janosh/matbench-discovery/blob/main/matbench_discovery/data.py
    """
    try:
        return obj.as_dict()  # all MSONable objects implement as_dict()
    except AttributeError:
        return None


@requires(mbd_installed, message="Requires `matbench_discovery` to be installed")
class MaterialsDiscoveryReducer(JsonDFReducer):
    def __init__(
        self,
        benchmark_name: str,
        target_data_path: str,
        cse_data_path: str | None = None,
        elemental_references_path: str | None = None,
        index_name: str | None = None,
        corrections: Compatibility | None = MP2020Compatibility,
        max_error_threshold: float = 5.0,
        analyze_geo_opt: bool = True,
        geo_symprec: float = 1e-5,
    ):
        """
        Args:
            benchmark_name: Name of the benchmark, used for file naming
            target_data_path: Path to the target data JSON file
            cse_data_path: Path to the WBM computed structure entries JSON file
            elemental_references_path: Path to elemental energy references JSON file
            index_name: Optional name of the column to use as index
            corrections: Optional correction class to apply to all entries
            max_error_threshold: Maximum allowed mean absolute formation energy per atom error threshold
            analyze_geo_opt: Whether to analyze geometry of relaxed structures and compute RMSD
            geo_symprec: Symmetry precision of moyopy.
        """
        index_name = index_name or str(Key.mat_id)
        self._corrections = corrections
        self._max_error_threshold = max_error_threshold
        self._elemental_references_path = elemental_references_path
        self._cse_data_path = cse_data_path
        self._analyze_geo_opt = analyze_geo_opt
        self._geo_symprec = geo_symprec
        super().__init__(
            benchmark_name=benchmark_name,
            target_data_path=target_data_path,
            index_name=index_name,
        )

    @property
    def runner_type(self) -> type[RelaxationRunner]:
        """The runner type this reducer is associated with."""
        return RelaxationRunner

    @staticmethod
    def load_targets(path: str, index_name: str | None) -> pd.DataFrame:
        """Load target data from a JSON file into a pandas DataFrame.

        Args:
            path: Path to the target JSON file
            index_name: Optional name of the column to use as index

        Returns:
            DataFrame containing the target data, sorted by index
        """
        df_wbm = pd.read_csv(path)
        if index_name is not None:
            df_wbm = df_wbm.set_index(index_name)
        return df_wbm.sort_index()

    @staticmethod
    def _load_elemental_ref_energies(
        elemental_references_path: str,
    ) -> dict[str, float]:
        elem_ref_entries = (
            pd.read_json(elemental_references_path, typ="series")
            .map(ComputedEntry.from_dict)
            .to_dict()
        )
        elemental_ref_energies = {
            elem: entry.energy_per_atom for elem, entry in elem_ref_entries.items()
        }
        return elemental_ref_energies

    @staticmethod
    def _load_computed_structure_entries(
        cse_data_path: str, results: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Convert prediction results to computed structure entries with updated energies and structures.

        Returns:
            DataFrame of computed structure entries indexed by material IDs
        """
        # WBM DF used to obtain CSEs for corrections and phase diagram input
        df_wbm_cse = pd.DataFrame(pd.read_json(cse_data_path, lines=True)).set_index(
            Key.mat_id
        )
        df_wbm_cse = df_wbm_cse.sort_index()

        df_wbm_cse[Key.computed_structure_entry] = [
            ComputedStructureEntry.from_dict(dct)
            for dct in tqdm(
                df_wbm_cse[Key.computed_structure_entry], desc="Creating pmg CSEs"
            )
        ]
        # transfer energies and relaxed structures to ComputedStructureEntries in order to apply corrections.
        # corrections applied below are structure-dependent (for oxides and sulfides)
        df_result_cse: list[dict[str, str | ComputedStructureEntry]] = []
        for mat_id in tqdm(
            results.index,
            desc="Converting predicted structures and energies to computed structure entries",
        ):
            cse = df_wbm_cse.loc[mat_id, Key.computed_structure_entry].copy()
            atoms = decode(results.loc[mat_id, "atoms"])
            structure = AseAtomsAdaptor.get_structure(atoms)
            energy = results.loc[mat_id, "energy"]
            cse._energy = energy
            cse._structure = structure
            df_result_cse.append(
                {Key.mat_id: mat_id, Key.computed_structure_entry: cse}
            )
        df_result_cse = pd.DataFrame(df_result_cse).set_index(Key.mat_id)
        return df_result_cse, df_wbm_cse

    def _apply_corrections(
        self, computed_structure_entries: list[ComputedStructureEntry]
    ) -> None:
        """Apply compatibility corrections to computed structure entries.

        Args:
            computed_structure_entries: List of ComputedStructureEntry objects to apply corrections to

        Raises:
            ValueError: If not all entries were successfully processed after applying corrections
        """
        if self._corrections is not None:
            processed = self._corrections.process_entries(
                computed_structure_entries, verbose=True, clean=True
            )
            if len(processed) != len(computed_structure_entries):
                raise ValueError(
                    f"not all entries processed: {len(processed)=} {len(computed_structure_entries)=}"
                )

    def _analyze_relaxed_geometry(
        self,
        pred_structures: dict[str, Structure],
        target_structures: dict[str, Structure],
    ) -> dict[str, float]:
        """Analyze geometry of relaxed structures and calculate RMSD wrt to the target structures.

        Args:
            pred_structures: Dictionary mapping material IDs to predicted Structure objects
            target_structures: Dictionary mapping material IDs to target Structure objects

        Returns:
            Dictionary containing geometric analysis metrics
        """
        df_symm_pred = symmetry.get_sym_info_from_structs(
            pred_structures,
            symprec=self._geo_symprec,
        )
        df_symm_target = symmetry.get_sym_info_from_structs(
            target_structures,
            symprec=self._geo_symprec,
        )
        df_geo_analysis = symmetry.pred_vs_ref_struct_symmetry(
            df_symm_pred,
            df_symm_target,
            pred_structures,
            target_structures,
        )
        return df_geo_analysis

    def join_results(self, results_dir: str, glob_pattern: str) -> pd.DataFrame:
        """Join results from multiple relaxation JSON files into a single DataFrame.

        Joins results for relaxed energy, applies compatibility corrections, and computes formation energy
        w.r.t to MP reference structures in MatBench Discovery

        Args:
            results_dir: Directory containing result files
            glob_pattern: Pattern to match result files

        Returns:
            Combined DataFrame containing all results
        """
        results = pd.concat(
            [
                pd.read_json(f).set_index("sid")
                for f in tqdm(
                    sorted(glob(os.path.join(results_dir, glob_pattern))),
                    desc=f"Loading results from {results_dir}",
                )
            ]
        ).sort_index()
        results.index.name = Key.mat_id

        df_cse_pred, df_cse_target = self._load_computed_structure_entries(
            self._cse_data_path, results
        )
        self._apply_corrections(df_cse_pred[Key.computed_structure_entry].tolist())

        # compute formation energy per atom
        elemental_ref_energies = self._load_elemental_ref_energies(
            self._elemental_references_path
        )
        results["e_form_per_atom"] = [
            calc_energy_from_e_refs(cse, ref_energies=elemental_ref_energies)
            for cse in tqdm(
                df_cse_pred[Key.computed_structure_entry],
                total=len(results),
                desc="Computing formation energies",
            )
        ]

        if self._analyze_geo_opt:
            pred_structures = {
                mat_id: cse.structure
                for mat_id, cse in df_cse_pred[Key.computed_structure_entry].items()
            }
            target_structures = {
                mat_id: cse.structure
                for mat_id, cse in df_cse_target[Key.computed_structure_entry].items()
            }
            df_geo_analysis = self._analyze_relaxed_geometry(
                pred_structures, target_structures
            )
            results = results.join(df_geo_analysis)

        return results

    def save_results(self, results: pd.DataFrame, results_dir: str) -> None:
        """Save joined results to a single file

        Saves the results in two formats:
        1. CSV file containing only numerical data
        2. JSON file containing all data including relaxed structures

        Args:
            results: DataFrame containing the prediction results
            results_dir: Directory path where result files will be saved
        """
        # save only numerical results
        results.select_dtypes("number").to_csv(
            os.path.join(results_dir, f"{self.benchmark_name}_results.csv.gz")
        )

        # save results including relaxed structures
        results.reset_index().to_json(
            os.path.join(results_dir, f"{self.benchmark_name}_results.json.gz"),
            default_handler=as_dict_handler,
            orient="records",
            lines=True,
        )

    def compute_metrics(self, results: pd.DataFrame, run_name: str) -> pd.DataFrame:
        """Compute Matbench discovery metrics for relaxed energy and structure predictions.

        Args:
            results: DataFrame containing prediction results with energy values
            run_name: Identifier for the current evaluation run

        Returns:
            DataFrame containing computed metrics for different material subsets
        """
        df_wbm = self.target_data
        df_wbm[[*results]] = results.round(4)

        # remove bad outlier predictions
        bad_mask = (
            abs(df_wbm["e_form_per_atom"] - df_wbm[MbdKey.e_form_dft])
            > self._max_error_threshold
        )
        df_wbm.loc[bad_mask, "e_form_per_atom"] = pd.NA
        n_preds, n_bad = len(df_wbm.dropna()), sum(bad_mask)
        logging.info(
            f"{n_bad:,} of {n_preds:,} unrealistic predictions with formation energy error "
            f"> {self._max_error_threshold} eV/atom"
        )

        # compute the predicted energy above hull
        e_above_hull_pred = (
            df_wbm[MbdKey.each_true]
            + df_wbm["e_form_per_atom"]
            - df_wbm[MbdKey.e_form_dft]
        )

        df_uniq_proto = df_wbm[df_wbm[MbdKey.uniq_proto]]
        e_above_hull_pred_uniq_proto = (
            df_uniq_proto[MbdKey.each_true]
            + df_uniq_proto["e_form_per_atom"]
            - df_uniq_proto[MbdKey.e_form_dft]
        )

        # now compute metrics
        metrics: dict[str, float] = stable_metrics(
            df_wbm[MbdKey.each_true], e_above_hull_pred, fillna=True
        )

        metrics_uniq_proto: dict[str, float] = stable_metrics(
            df_uniq_proto[MbdKey.each_true], e_above_hull_pred_uniq_proto, fillna=True
        )

        # Get the 10,000 most stable materials based on predicted energy above hull
        most_stable_10k = e_above_hull_pred_uniq_proto.nsmallest(10_000)
        metrics_most_stable_10k = stable_metrics(
            df_wbm[MbdKey.each_true].loc[most_stable_10k.index],
            most_stable_10k,
            fillna=True,
        )

        if self._analyze_geo_opt:
            metrics.update(calc_geo_opt_metrics(results))
            metrics_uniq_proto.update(
                calc_geo_opt_metrics(results[df_wbm[MbdKey.uniq_proto]])
            )
            metrics_most_stable_10k.update(
                calc_geo_opt_metrics(results.loc[most_stable_10k.index])
            )

        all_metrics = pd.DataFrame(
            [metrics, metrics_uniq_proto, metrics_most_stable_10k],
            index=["full", "uniq-proto", "most-stable-10k"],
        )

        return all_metrics

    def log_metrics(self, metrics: pd.DataFrame, run_name: str) -> None:
        """Log metrics to the configured logger if available.

        Args:
            metrics: DataFrame containing the computed metrics
            run_name: Name of the current run
        """
        # drop these columns for cleaner logging
        metrics = metrics.drop(columns=["TP", "FP", "TN", "FN"])
        if self.logger is not None:
            for split in metrics.index:  # log each MBD split into a different table
                split_metrics = pd.DataFrame(
                    data=[metrics.loc[split]],
                    index=[run_name],
                )
                self.logger.log_dataframe(
                    name=f"{self.benchmark_name}-{split}",
                    dataframe=split_metrics.reset_index(),
                )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True

    def load_state(self, checkpoint_location: str | None) -> None:
        pass



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from contextlib import contextmanager
from copy import deepcopy
from typing import TYPE_CHECKING

import hydra
import torch
from omegaconf import DictConfig

from fairchem.core.common.registry import registry
from fairchem.core.common.utils import load_state_dict, match_state_dict

if TYPE_CHECKING:
    from fairchem.core.units.mlip_unit.api.inference import MLIPInferenceCheckpoint
    from fairchem.core.units.mlip_unit.mlip_unit import Task


def get_backbone_class_from_checkpoint(
    checkpoint: MLIPInferenceCheckpoint,
) -> type:
    """Extract the backbone class from a checkpoint's config."""
    backbone_config = checkpoint.model_config.get("backbone", {})
    backbone_model_name = backbone_config.get("model")

    if backbone_model_name is None:
        raise ValueError("Cannot determine backbone class from checkpoint config")

    return registry.get_model_class(backbone_model_name)


def load_inference_model(
    checkpoint_location: str,
    overrides: dict | None = None,
    use_ema: bool = False,
    return_checkpoint: bool = True,
    strict: bool = True,
    preloaded_checkpoint: MLIPInferenceCheckpoint | None = None,
) -> tuple[torch.nn.Module, MLIPInferenceCheckpoint] | torch.nn.Module:
    if preloaded_checkpoint is not None:
        checkpoint = preloaded_checkpoint
    else:
        checkpoint = torch.load(
            checkpoint_location, map_location="cpu", weights_only=False
        )

    if overrides is not None:
        checkpoint.model_config = update_configs(checkpoint.model_config, overrides)

    model = hydra.utils.instantiate(checkpoint.model_config)
    if use_ema:
        model = torch.optim.swa_utils.AveragedModel(model)
        model_dict = model.state_dict()
        ema_state_dict = checkpoint.ema_state_dict

        n_averaged = ema_state_dict["n_averaged"]
        del model_dict["n_averaged"]
        del ema_state_dict["n_averaged"]

        matched_dict = match_state_dict(model_dict, ema_state_dict)

        matched_dict["n_averaged"] = n_averaged

        load_state_dict(model, matched_dict, strict=strict)
    else:
        load_state_dict(model, checkpoint.model_state_dict, strict=strict)

    return (model, checkpoint) if return_checkpoint is True else model


def load_tasks(checkpoint_location: str) -> list[Task]:
    """
    Load tasks from a checkpoint file.

    Args:
        checkpoint_location (str): Path to the checkpoint file.

    Returns:
        list[Task]: A list of instantiated Task objects from the checkpoint's tasks_config.
    """
    checkpoint: MLIPInferenceCheckpoint = torch.load(
        checkpoint_location, map_location="cpu", weights_only=False
    )
    return [
        hydra.utils.instantiate(task_config) for task_config in checkpoint.tasks_config
    ]


@contextmanager
def tf32_context_manager():
    # Store the original settings
    original_allow_tf32_matmul = torch.backends.cuda.matmul.allow_tf32
    original_allow_tf32_cudnn = torch.backends.cudnn.allow_tf32
    original_float32_matmul_precision = torch.get_float32_matmul_precision()
    try:
        # Set the desired settings
        torch.backends.cuda.matmul.allow_tf32 = True
        torch.backends.cudnn.allow_tf32 = True
        torch.set_float32_matmul_precision("high")
        yield
    finally:
        # Revert to the original settings
        torch.backends.cuda.matmul.allow_tf32 = original_allow_tf32_matmul
        torch.backends.cudnn.allow_tf32 = original_allow_tf32_cudnn
        torch.set_float32_matmul_precision(original_float32_matmul_precision)


def update_configs(original_config, new_config):
    updated_config = deepcopy(original_config)
    for k, v in new_config.items():
        is_dict_config = (isinstance(v, (dict, DictConfig))) and (
            isinstance(updated_config[k], (dict, DictConfig))
        )
        if is_dict_config and k in updated_config:
            updated_config[k] = update_configs(updated_config[k], v)
        else:
            updated_config[k] = v
    return updated_config



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from ._benchmark_reducer import JsonDFReducer
from ._single.adsorbml_reducer import AdsorbMLReducer
from ._single.adsorption_reducer import AdsorptionReducer
from ._single.kappa_reducer import Kappa103Reducer
from ._single.materials_discovery_reducer import MaterialsDiscoveryReducer
from ._single.nvemd_reducer import NVEMDReducer
from ._single.omc_polymorph_reducer import OMCPolymorphReducer
from ._single.omol_reducer import OMolReducer
from ._single.uma_speed_benchmark import InferenceBenchRunner

__all__ = [
    "JsonDFReducer",
    "AdsorbMLReducer",
    "AdsorptionReducer",
    "Kappa103Reducer",
    "MaterialsDiscoveryReducer",
    "NVEMDReducer",
    "OMCPolymorphReducer",
    "OMolReducer",
    "InferenceBenchRunner",
]



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import torch

from fairchem.core.datasets.atomic_data import AtomicData, atomicdata_list_to_batch


class MTCollater:
    def __init__(
        self,
        task_config,
        exclude_keys,
        otf_graph: bool = False,  # Deprecated, do not use!
    ) -> None:
        self.exclude_keys = exclude_keys
        self.task_config = task_config
        self.dataset_task_map = self._create_dataset_task_map(task_config)

    def __call__(self, data_list: list[AtomicData]) -> AtomicData:
        batch = self.data_list_collater(
            data_list,
            dataset_task_map=self.dataset_task_map,
            exclude_keys=self.exclude_keys,
        )
        return batch

    def data_list_collater(
        self,
        data_list: list[AtomicData],
        dataset_task_map: dict,
        exclude_keys: list,
    ) -> AtomicData:
        data_list = self._add_missing_attr(data_list, dataset_task_map)

        return atomicdata_list_to_batch(data_list, exclude_keys=exclude_keys)

    # takes in the task config
    def _create_dataset_task_map(self, config):
        datasets_task_map = {}
        for task, task_dict in config.items():
            for dataset in task_dict["datasets"]:
                if dataset not in datasets_task_map:
                    datasets_task_map[dataset] = {}
                datasets_task_map[dataset][task] = {"level": task_dict["level"]}
        return datasets_task_map

    def _add_missing_attr(self, data_list, dataset_task_map):
        """
        add missing data object attributes as inf
        """
        # find all tasks in a given batch
        # and collect all necessary info
        datasets_in_batch_to_task_configs = {}
        tasks_to_report_on = []

        for data in data_list:
            if data.dataset not in datasets_in_batch_to_task_configs:
                datasets_in_batch_to_task_configs[data.dataset] = dataset_task_map[
                    data.dataset
                ]
                for task, task_config in datasets_in_batch_to_task_configs[
                    data.dataset
                ].items():
                    # if this is the first time we have seen this kind of dataset
                    # record its out spec
                    if (
                        "out_spec"
                        not in datasets_in_batch_to_task_configs[data.dataset][task]
                    ):
                        if task_config["level"] == "system":
                            dim = list(getattr(data, task).shape)
                        elif task_config["level"] == "atom":
                            dim = list(getattr(data, task).shape[1:])
                        else:
                            raise ValueError(
                                f"task level must be either system or atom, found {task_config['level']}"
                            )
                        datasets_in_batch_to_task_configs[data.dataset][task][
                            "out_spec"
                        ] = {
                            "dim": dim,
                            "dtype": str(getattr(data, task).dtype),
                        }

                    # make sure all data in this batch has a valid output for this task
                    tasks_to_report_on.append(task)

        # tasks info
        set_all_tasks = True  # DDP breaks if we try and be clever here :(
        if set_all_tasks:
            task_config = self.task_config
            tasks_to_report_on = list(self.task_config)
        else:
            task_config = {
                task: val
                for _, td in datasets_in_batch_to_task_configs.items()
                for task, val in td.items()
            }

        # find missing tasks in the batch
        missing_tasks = {}
        for dataset, task_dict in datasets_in_batch_to_task_configs.items():
            missing_tasks[dataset] = list(
                set(tasks_to_report_on).difference(set(task_dict.keys()))
            )

        # create a copy before assigning attributes to inf
        data_list = [data.clone() for data in data_list]

        # set missing attributes to inf for all data objects in the batch
        # according to level, output dim, and dtype
        for data in data_list:
            assert (
                len(data) == 1
            ), "data must contain a single element, can not batch batches"
            assert len(data.dataset) == 1, "dataset name must be a single string"
            for task in missing_tasks[data.dataset[0]]:
                dim = task_config[task]["out_spec"]["dim"]
                dtype = getattr(torch, task_config[task]["out_spec"]["dtype"])
                if task_config[task]["level"] == "atom":
                    dim = [data.natoms] + dim
                setattr(data, task, torch.full(dim, torch.inf, dtype=dtype))
        return data_list



"""
Copyright (c) Meta, Inc. and its affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import torch
import torch.nn as nn


# Different encodings for the atom distance embeddings
class GaussianSmearing(torch.nn.Module):
    def __init__(
        self,
        start: float = -5.0,
        stop: float = 5.0,
        num_gaussians: int = 50,
        basis_width_scalar: float = 1.0,
    ) -> None:
        super().__init__()
        self.num_output = num_gaussians
        offset = torch.linspace(start, stop, num_gaussians)
        self.coeff = -0.5 / (basis_width_scalar * (offset[1] - offset[0])).item() ** 2
        self.register_buffer("offset", offset)

    def forward(self, dist) -> torch.Tensor:
        dist = dist.view(-1, 1) - self.offset.view(1, -1)  # type: ignore
        return torch.exp(self.coeff * torch.pow(dist, 2))


class SigmoidSmearing(torch.nn.Module):
    def __init__(
        self, start=-5.0, stop=5.0, num_sigmoid=50, basis_width_scalar=1.0
    ) -> None:
        super().__init__()
        self.num_output = num_sigmoid
        offset = torch.linspace(start, stop, num_sigmoid)
        self.coeff = (basis_width_scalar / (offset[1] - offset[0])).item()
        self.register_buffer("offset", offset)

    def forward(self, dist) -> torch.Tensor:
        exp_dist = self.coeff * (dist.view(-1, 1) - self.offset.view(1, -1))  # type: ignore
        return torch.sigmoid(exp_dist)


class LinearSigmoidSmearing(torch.nn.Module):
    def __init__(
        self,
        start: float = -5.0,
        stop: float = 5.0,
        num_sigmoid: int = 50,
        basis_width_scalar: float = 1.0,
    ) -> None:
        super().__init__()
        self.num_output = num_sigmoid
        offset = torch.linspace(start, stop, num_sigmoid)
        self.coeff = (basis_width_scalar / (offset[1] - offset[0])).item()
        self.register_buffer("offset", offset)

    def forward(self, dist) -> torch.Tensor:
        exp_dist = self.coeff * (dist.view(-1, 1) - self.offset.view(1, -1))  # type: ignore
        return torch.sigmoid(exp_dist) + 0.001 * exp_dist


class SiLUSmearing(torch.nn.Module):
    def __init__(
        self,
        start: float = -5.0,
        stop: float = 5.0,
        num_output: int = 50,
        basis_width_scalar: float = 1.0,
    ) -> None:
        super().__init__()
        self.num_output = num_output
        self.fc1 = nn.Linear(2, num_output)
        self.act = nn.SiLU()

    def forward(self, dist):
        x_dist = dist.view(-1, 1)
        x_dist = torch.cat([x_dist, torch.ones_like(x_dist)], dim=1)
        return self.act(self.fc1(x_dist))



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import copy
import logging
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from typing import TYPE_CHECKING, Sequence

import hydra
import torch
from torch import nn

from fairchem.core.common.registry import registry
from fairchem.core.common.utils import (
    load_model_and_weights_from_checkpoint,
)

if TYPE_CHECKING:
    from ase import Atoms

    from fairchem.core.datasets.atomic_data import AtomicData
    from fairchem.core.units.mlip_unit.api.inference import InferenceSettings
    from fairchem.core.units.mlip_unit.mlip_unit import Task


def _get_dataset_to_tasks_map(tasks: Sequence[Task]) -> dict[str, list[Task]]:
    """Create a mapping from dataset names to their associated tasks."""
    dset_to_tasks_map = defaultdict(list)
    for task in tasks:
        for dataset_name in task.datasets:
            dset_to_tasks_map[dataset_name].append(task)
    return dict(dset_to_tasks_map)


class HeadInterface(metaclass=ABCMeta):
    @property
    def use_amp(self):
        return False

    @abstractmethod
    def forward(
        self, data: AtomicData, emb: dict[str, torch.Tensor]
    ) -> dict[str, torch.Tensor]:
        """Head forward.

        Arguments
        ---------
        data: AtomicData
            Atomic systems as input
        emb: dict[str->torch.Tensor]
            Embeddings of the input as generated by the backbone

        Returns
        -------
        outputs: dict[str->torch.Tensor]
            Return one or more targets generated by this head
        """
        return


class BackboneInterface(metaclass=ABCMeta):
    @abstractmethod
    def forward(self, data: AtomicData) -> dict[str, torch.Tensor]:
        """Backbone forward.

        Arguments
        ---------
        data: AtomicData
            Atomic systems as input

        Returns
        -------
        embedding: dict[str->torch.Tensor]
            Return backbone embeddings for the given input
        """
        return

    @classmethod
    @abstractmethod
    def build_inference_settings(cls, settings: InferenceSettings) -> dict:
        """
        Build config overrides from inference settings.

        Called BEFORE model instantiation. Override in subclasses to handle
        model-specific settings like activation_checkpointing.

        Can also perform validation of settings and raise errors if incompatible.

        Returns:
            Dict of backbone config key-value pairs to override.
        """
        return {}

    @abstractmethod
    def validate_tasks(self, dataset_to_tasks: dict[str, list]) -> None:
        """Validate that task datasets are compatible with this backbone."""
        pass  # noqa

    @abstractmethod
    def prepare_for_inference(self, data: AtomicData, settings: InferenceSettings):
        """Prepare backbone for inference. Called once on first prediction.

        Returns:
            self or a new backbone if replacement occurred (e.g., after MOLE merge).
        """
        return self

    @abstractmethod
    def on_predict_check(self, data: AtomicData) -> None:
        """Called before each prediction for any per-prediction checks."""
        pass  # noqa

    @abstractmethod
    def validate_atoms_data(self, atoms: Atoms, task_name: str) -> None:
        """Validate and set defaults for calculator input data.

        Models should set appropriate defaults for atoms.info (e.g., charge, spin)
        and validate that required fields are present and within valid ranges.
        """
        pass  # noqa


@registry.register_model("hydra")
class HydraModel(nn.Module):
    def __init__(
        self,
        backbone: dict | None = None,
        heads: dict | None = None,
        finetune_config: dict | None = None,
        otf_graph: bool = True,
        pass_through_head_outputs: bool = False,
        freeze_backbone: bool = False,
    ):
        super().__init__()
        self.device = None
        self.otf_graph = otf_graph
        # This is required for hydras with models that have multiple outputs per head, since we will deprecate
        # the old config system at some point, this will prevent the need to make major modifications to the trainer
        # because they all expect the name of the outputs directly instead of the head_name.property_name
        self.pass_through_head_outputs = pass_through_head_outputs

        # if finetune_config is provided, then attempt to load the model from the given finetune checkpoint
        starting_model = None
        if finetune_config is not None:
            # Make it hard to sneak more fields into finetuneconfig
            assert (
                len(set(finetune_config.keys()) - {"starting_checkpoint", "override"})
                == 0
            )
            starting_model: HydraModel = load_model_and_weights_from_checkpoint(
                finetune_config["starting_checkpoint"]
            )
            logging.info(
                f"Found and loaded fine-tuning checkpoint: {finetune_config['starting_checkpoint']} (Note we are NOT loading the training state from this checkpoint, only parts of the model and weights)"
            )
            assert isinstance(
                starting_model, HydraModel
            ), "Can only finetune starting from other hydra models!"
            # TODO this is a bit hacky to overrride attrs in the backbone
            if "override" in finetune_config:
                for key, value in finetune_config["override"].items():
                    setattr(starting_model.backbone, key, value)

        if backbone is not None:
            backbone = copy.deepcopy(backbone)
            backbone_model_name = backbone.pop("model")
            self.backbone: BackboneInterface = registry.get_model_class(
                backbone_model_name
            )(
                **backbone,
            )
        elif starting_model is not None:
            self.backbone = starting_model.backbone
            logging.info(
                f"User did not specify a backbone, using the backbone from the starting checkpoint {self.backbone}"
            )
        else:
            raise RuntimeError(
                "Backbone not specified and not found in the starting checkpoint"
            )

        if freeze_backbone:
            for param in self.backbone.parameters():
                param.requires_grad = False

        if heads is not None:
            heads = copy.deepcopy(heads)
            # Iterate through outputs_cfg and create heads
            self.output_heads: dict[str, HeadInterface] = {}

            head_names_sorted = sorted(heads.keys())
            assert len(set(head_names_sorted)) == len(
                head_names_sorted
            ), "Head names must be unique!"
            for head_name in head_names_sorted:
                head_config = heads[head_name]
                if "module" not in head_config:
                    raise ValueError(
                        f"{head_name} head does not specify module to use for the head"
                    )

                module_name = head_config.pop("module")
                self.output_heads[head_name] = registry.get_model_class(module_name)(
                    self.backbone,
                    **head_config,
                )

            self.output_heads = torch.nn.ModuleDict(self.output_heads)
        elif starting_model is not None:
            self.output_heads = starting_model.output_heads
            logging.info(
                f"User did not specify heads, using the output heads from the starting checkpoint {self.output_heads}"
            )
        else:
            raise RuntimeError(
                "Heads not specified and not found in the starting checkpoint"
            )

    def forward(self, data: AtomicData):
        # lazily get device from input to use with amp, at least one input must be a tensor to figure out it's device
        if not self.device:
            device_from_tensors = {
                x.device.type for x in data.values() if isinstance(x, torch.Tensor)
            }
            assert (
                len(device_from_tensors) == 1
            ), f"all inputs must be on the same device, found the following devices {device_from_tensors}"
            self.device = device_from_tensors.pop()

        emb = self.backbone(data)
        # Predict all output properties for all structures in the batch for now.
        out = {}
        for k in self.output_heads:
            with torch.autocast(
                device_type=self.device, enabled=self.output_heads[k].use_amp
            ):
                if self.pass_through_head_outputs:
                    out.update(self.output_heads[k](data, emb))
                else:
                    out[k] = self.output_heads[k](data, emb)

        return out

    @property
    def direct_forces(self) -> bool:
        """
        Whether this model uses direct force prediction.
        """
        return getattr(self.backbone, "direct_forces", False)

    def prepare_for_inference(
        self, data: AtomicData, settings: InferenceSettings
    ) -> None:
        """
        Prepare model for inference. Called once on first prediction.

        Delegates to backbone and each head that implements prepare_for_inference.
        """
        need_eval = False
        # Backbone may return a new backbone (e.g., after MOLE merge)
        new_backbone = self.backbone.prepare_for_inference(data, settings)
        if new_backbone is not self.backbone:
            self.backbone = new_backbone
            # Re-set eval mode after backbone replacement
            need_eval = True

        # Let each head prepare itself if needed (returns new head or self)
        new_output_heads = torch.nn.ModuleDict()
        heads_changed = False
        for head_name, head in self.output_heads.items():
            if hasattr(head, "prepare_for_inference"):
                new_head = head.prepare_for_inference(data, settings)
                new_output_heads[head_name] = new_head
                if new_head is not head:
                    heads_changed = True
            else:
                new_output_heads[head_name] = head

        if heads_changed:
            self.output_heads = new_output_heads
            need_eval = True

        if need_eval:
            self.eval()

    def on_predict_check(self, data: AtomicData) -> None:
        """
        Called before each prediction for any per-prediction checks.
        """
        self.backbone.on_predict_check(data)

    def validate_atoms_data(self, atoms: Atoms, task_name: str) -> None:
        """
        Validate and set defaults for calculator input data.
        """
        self.backbone.validate_atoms_data(atoms, task_name)

    def setup_tasks(self, tasks_config: list) -> None:
        """
        Setup tasks from checkpoint config.

        Args:
            tasks_config: List of task configurations from checkpoint
        """
        tasks = [hydra.utils.instantiate(task_config) for task_config in tasks_config]
        self.tasks = {t.name: t for t in tasks}
        self._dataset_to_tasks = _get_dataset_to_tasks_map(tasks)

        # Let backbone validate tasks
        self.backbone.validate_tasks(self._dataset_to_tasks)

    @property
    def dataset_to_tasks(self) -> dict[str, list]:
        """
        Mapping from dataset names to their associated tasks.
        """
        if not hasattr(self, "_dataset_to_tasks"):
            raise RuntimeError(
                "setup_tasks() must be called before accessing dataset_to_tasks"
            )
        return self._dataset_to_tasks


class HydraModelV2(nn.Module):
    def __init__(
        self,
        backbone: BackboneInterface,
        heads: dict[str, HeadInterface],
        freeze_backbone: bool = False,
    ):
        super().__init__()
        self.backbone = backbone
        self.output_heads = torch.nn.ModuleDict(heads)
        self.device = None
        if freeze_backbone:
            for param in self.backbone.parameters():
                param.requires_grad = False

    def forward(self, data):
        # lazily get device from input to use with amp, at least one input must be a tensor to figure out it's device
        if not self.device:
            device_from_tensors = {
                x.device.type for x in data.values() if isinstance(x, torch.Tensor)
            }
            assert (
                len(device_from_tensors) == 1
            ), f"all inputs must be on the same device, found the following devices {device_from_tensors}"
            self.device = device_from_tensors.pop()

        emb = self.backbone(data)
        # Predict all output properties for all structures in the batch for now.
        out = {}
        for k in self.output_heads:
            with torch.autocast(
                device_type=self.device, enabled=self.output_heads[k].use_amp
            ):
                out[k] = self.output_heads[k](data, emb)
        return out

    @property
    def direct_forces(self) -> bool:
        """
        Whether this model uses direct force prediction.
        """
        return getattr(self.backbone, "direct_forces", False)

    def prepare_for_inference(
        self, data: AtomicData, settings: InferenceSettings
    ) -> None:
        """
        Prepare model for inference. Called once on first prediction.
        """
        data_clone = data.clone()
        need_eval = False
        # Backbone may return a new backbone (e.g., after MOLE merge)
        new_backbone = self.backbone.prepare_for_inference(data_clone, settings)
        if new_backbone is not self.backbone:
            self.backbone = new_backbone
            need_eval = True

        new_output_heads = torch.nn.ModuleDict()
        heads_changed = False
        for head_name, head in self.output_heads.items():
            if hasattr(head, "prepare_for_inference"):
                new_head = head.prepare_for_inference(data_clone, settings)
                new_output_heads[head_name] = new_head
                if new_head is not head:
                    heads_changed = True
            else:
                new_output_heads[head_name] = head

        if heads_changed:
            self.output_heads = new_output_heads
            need_eval = True

        if need_eval:
            self.eval()

    def on_predict_check(self, data: AtomicData) -> None:
        """
        Called before each prediction for any per-prediction checks.
        """
        self.backbone.on_predict_check(data)

    def validate_atoms_data(self, atoms: Atoms, task_name: str) -> None:
        """
        Validate and set defaults for calculator input data.
        """
        self.backbone.validate_atoms_data(atoms, task_name)

    def setup_tasks(self, tasks_config: list) -> None:
        """
        Setup tasks from checkpoint config.

        Args:
            tasks_config: List of task configurations from checkpoint
        """
        tasks = [hydra.utils.instantiate(task_config) for task_config in tasks_config]
        self.tasks = {t.name: t for t in tasks}
        self._dataset_to_tasks = _get_dataset_to_tasks_map(tasks)

        # Let backbone validate tasks
        self.backbone.validate_tasks(self._dataset_to_tasks)

    @property
    def dataset_to_tasks(self) -> dict[str, list]:
        """
        Mapping from dataset names to their associated tasks.
        """
        if not hasattr(self, "_dataset_to_tasks"):
            raise RuntimeError(
                "setup_tasks() must be called before accessing dataset_to_tasks"
            )
        return self._dataset_to_tasks



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import clusterscope
import hydra
import ray
import torch.distributed as dist
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy
from torch.distributed.elastic.utils.distributed import get_free_port

from fairchem.core.common import gp_utils
from fairchem.core.common.distutils import (
    assign_device_for_local_rank,
    setup_env_local_multi_gpu,
)
from fairchem.core.common.utils import setup_env_vars
from fairchem.core.components.runner import Runner
from fairchem.core.launchers.api import DeviceType
from fairchem.core.launchers.cluster.ray_cluster import RayCluster

if TYPE_CHECKING:
    from omegaconf import DictConfig

    from fairchem.core.launchers.api import SchedulerConfig, SlurmConfig


@ray.remote
class SPMDWorker:
    def __init__(
        self,
        job_config: DictConfig,
        runner_config: DictConfig,
        worker_id: int,
        world_size: int,
        device: str,
        gp_size: int | None = None,
        master_addr: str | None = None,
        master_port: int | None = None,
    ):
        self.runner_config = runner_config
        # master address and port is not passed in, initialize it here
        self.master_address = (
            ray.util.get_node_ip_address() if master_addr is None else master_addr
        )
        self.master_port = get_free_port() if master_port is None else master_port
        self.worker_id = worker_id
        self.device = device
        self.gp_size = gp_size
        self.world_size = world_size
        self.job_config = job_config
        setup_env_vars()
        self.distributed_setup = False

    def _distributed_setup(
        self,
        worker_id: int,
        world_size: int,
        master_address: str,
        master_port: int,
        device: str,
        gp_size: int | None,
    ):
        setup_env_local_multi_gpu(worker_id, master_port, master_address)
        assign_device_for_local_rank(device == "cpu", 0)
        backend = "gloo" if device == "cpu" else "nccl"
        dist.init_process_group(
            backend=backend,
            rank=worker_id,
            world_size=world_size,
        )
        if gp_size is not None:
            gp_utils.setup_graph_parallel_groups(gp_size, backend)

    def get_master_address_and_port(self):
        return (self.master_address, self.master_port)

    def run(self):
        if not self.distributed_setup:
            # initialize distributed environment
            self._distributed_setup(
                worker_id=self.worker_id,
                world_size=self.world_size,
                master_address=self.master_address,
                master_port=self.master_port,
                device=self.device,
                gp_size=self.gp_size,
            )
            self.runner: Runner = hydra.utils.instantiate(self.runner_config)
            self.runner.job_config = self.job_config
            self.distributed_setup = True
        self.runner.run()


class SPMDController(Runner):
    # this is equivalent to the fairchem SlurmSPMDProgram routine that runs the runner on every worker
    def __init__(self, job_config: DictConfig, runner_config: DictConfig):
        self.job_config = job_config
        self.runner_config = runner_config
        self.device = job_config.device_type.value
        self.world_size = (
            job_config.scheduler.num_nodes * job_config.scheduler.ranks_per_node
        )
        self.gp_group_size = job_config.graph_parallel_group_size
        self.ranks_per_node = job_config.scheduler.ranks_per_node
        self.num_nodes = job_config.scheduler.num_nodes
        num_gpus_per_group = (
            self.ranks_per_node if job_config.device_type == DeviceType.CUDA else 0
        )
        bundle_gpus = {
            "GPU": num_gpus_per_group,
            "CPU": self.ranks_per_node,
        }
        placement_groups = []
        # first create one placement group for each node
        for _ in range(self.num_nodes):
            pg = ray.util.placement_group([bundle_gpus], strategy="STRICT_PACK")
            placement_groups.append(pg)
        ray.get(pg.ready())  # Wait for each placement group to be scheduled

        logging.info(f"{len(placement_groups)} placement groups are ready")
        rank0_worker = SPMDWorker.options(
            num_gpus=1 if num_gpus_per_group > 0 else 0,
            scheduling_strategy=PlacementGroupSchedulingStrategy(
                placement_group=placement_groups[0],
                placement_group_bundle_index=0,  # Use the first (and only) bundle in the PG
                placement_group_capture_child_tasks=True,  # Ensure child tasks also run in this PG
            ),
        ).remote(
            self.job_config,
            self.runner_config,
            0,
            self.world_size,
            self.device,
            self.gp_group_size,
            None,
            None,
        )
        master_addr, master_port = ray.get(
            rank0_worker.get_master_address_and_port.remote()
        )
        logging.info(f"Started rank0 on {master_addr}:{master_port}")
        self.workers = [rank0_worker]

        # next place all ranks in order and pack them on placement groups
        # ie: rank0-7 -> placement group 0, 8->15 -> placement group 1 etc.
        for pg_idx, pg in enumerate(placement_groups):
            print(f"Launching workers for placement group {pg_idx} (Node {pg_idx})")

            for gpu_rank_on_node in range(self.ranks_per_node):
                if pg_idx == 0 and gpu_rank_on_node == 0:
                    continue
                # Each actor requests 1 GPU and uses the specific placement group
                actor = SPMDWorker.options(
                    num_gpus=1 if num_gpus_per_group > 0 else 0,
                    scheduling_strategy=PlacementGroupSchedulingStrategy(
                        placement_group=pg,
                        placement_group_bundle_index=0,  # Use the first (and only) bundle in the PG
                        placement_group_capture_child_tasks=True,  # Ensure child tasks also run in this PG
                    ),
                ).remote(
                    self.job_config,
                    self.runner_config,
                    pg_idx * self.ranks_per_node + gpu_rank_on_node,
                    self.world_size,
                    self.device,
                    self.gp_group_size,
                    master_addr,
                    master_port,
                )
                self.workers.append(actor)

    def run(self):
        logging.info("Running SPMDWrapper payload ...")
        futures = [w.run.remote() for w in self.workers]
        ray.get(futures)

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        pass

    def load_state(self, checkpoint_location: str | None) -> None:
        pass


def ray_entrypoint(runner_config: DictConfig, recursive_instantiate_runner: bool):
    runner = hydra.utils.instantiate(
        runner_config, _recursive_=recursive_instantiate_runner
    )
    runner.run()


def ray_on_slurm_launch(config: DictConfig, log_dir: str):
    scheduler_config: SchedulerConfig = config.job.scheduler
    slurm_config: SlurmConfig = scheduler_config.slurm
    cluster = RayCluster(log_dir=Path(log_dir))
    cluster_reqs = {
        "slurm_account": slurm_config.account,
        "slurm_qos": slurm_config.qos,
        "timeout_min": slurm_config.timeout_hr * 60,
        "mem_gb": slurm_config.mem_gb,
        "nodes": scheduler_config.num_nodes,
        "gpus_per_task": scheduler_config.ranks_per_node,
        "cpus_per_task": clusterscope.cpus(),  # need to request all the cpus on the node
        "tasks_per_node": 1,
    }
    cluster.start_head_and_workers(
        name=config.job.run_name,
        requirements=cluster_reqs,
        payload=ray_entrypoint,
        runner_config=config.runner,
        recursive_instantiate_runner=config.job.recursive_instantiate_runner,
    )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any

from omegaconf import DictConfig

from fairchem.core.components.utils import ManagedAttribute

if TYPE_CHECKING:
    from fairchem.core.components.runner import Runner


class Reducer(metaclass=ABCMeta):
    """Represents an abstraction over things that reduce the results written by a set of runner.

    Note:
        When running with the `fairchemv2` cli, the `job_config` and `runner_config` attributes are set at
        runtime to those given in the config file.

    Attributes:
        job_config (DictConfig): a managed attribute that gives access to the job config
        runner_config (DictConfig): a managed attributed that gives access to the calling runner config
    """

    job_config = ManagedAttribute(enforced_type=DictConfig)
    runner_config = ManagedAttribute(enforced_type=DictConfig)

    @property
    @abstractmethod
    def runner_type(self) -> type[Runner]:
        """The runner type this reducer is associated with."""
        raise NotImplementedError

    @abstractmethod
    def reduce(self) -> Any:
        """Use file pattern to reduce"""
        raise NotImplementedError

    @abstractmethod
    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        raise NotImplementedError

    @abstractmethod
    def load_state(self, checkpoint_location: str | None) -> None:
        raise NotImplementedError


class MockReducer(Reducer):
    """Used for testing"""

    @property
    def runner_type(self):
        from fairchem.core.components.runner import MockRunner

        return MockRunner

    def reduce(self) -> Any:
        runner_path = self.runner_config.pop("_target_")
        print(
            f"Reducing results from runner {runner_path} with args: "
            f"{', '.join(f'{k}: {v}' for k, v in self.runner_config.items())}"
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        pass

    def load_state(self, checkpoint_location: str | None) -> None:
        pass



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

Creates LMDB files with extracted graph features from provided *.extxyz files
for the S2EF task.
"""

from __future__ import annotations

import argparse
import glob
import multiprocessing as mp
import os
import pickle

import ase.io
import lmdb
import numpy as np
import torch
from tqdm import tqdm

from fairchem.core.preprocessing import AtomsToGraphs


def write_images_to_lmdb(mp_arg) -> None:
    a2g, db_path, samples, pid = mp_arg
    db = lmdb.open(
        db_path,
        map_size=1099511627776 * 2,
        subdir=False,
        meminit=False,
        map_async=True,
    )

    pbar = tqdm(
        total=len(samples),
        position=pid,
        desc="Preprocessing data into LMDBs",
    )
    idx = 0
    for sample in samples:
        ml_relaxed = ase.io.read(sample, "-1")
        data_object = a2g.convert(ml_relaxed)

        sid, _ = os.path.splitext(os.path.basename(sample))
        fid = -1
        # add atom tags
        data_object.tags = torch.LongTensor(ml_relaxed.get_tags())
        data_object.sid = int(sid)
        data_object.fid = fid

        txn = db.begin(write=True)
        txn.put(
            f"{idx}".encode("ascii"),
            pickle.dumps(data_object, protocol=-1),
        )
        txn.commit()
        idx += 1
        pbar.update(1)

    # Save count of objects in lmdb.
    txn = db.begin(write=True)
    txn.put("length".encode("ascii"), pickle.dumps(idx, protocol=-1))
    txn.commit()

    db.sync()
    db.close()


def main(args, split) -> None:
    systems = glob.glob(f"{eval(f'args.{split}')}/*.traj")

    systems_chunked = np.array_split(systems, args.num_workers)

    # Initialize feature extractor.
    a2g = AtomsToGraphs(
        max_neigh=50,
        radius=6,
        r_energy=False,
        r_forces=False,
        r_distances=False,
        r_fixed=True,
        r_edges=True,
    )

    # Create output directory if it doesn't exist.
    out_path = f"{args.out_path}_{split}"
    os.makedirs(out_path, exist_ok=True)

    # Initialize lmdb paths
    db_paths = [
        os.path.join(out_path, "data.%04d.lmdb" % i) for i in range(args.num_workers)
    ]

    pool = mp.Pool(args.num_workers)
    mp_args = [
        (
            a2g,
            db_paths[i],
            systems_chunked[i],
            i,
        )
        for i in range(args.num_workers)
    ]
    list(pool.imap(write_images_to_lmdb, mp_args))
    pool.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--id",
        required=True,
        help="Path to ID trajectories",
    )
    parser.add_argument(
        "--ood-ads",
        required=True,
        help="Path to OOD-Ads trajectories",
    )
    parser.add_argument(
        "--ood-cat",
        required=True,
        help="Path to OOD-Cat trajectories",
    )
    parser.add_argument(
        "--ood-both",
        required=True,
        help="Path to OOD-Both trajectories",
    )
    parser.add_argument(
        "--out-path",
        required=True,
        help="Directory to save extracted features. Will create if doesn't exist",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=1,
        help="No. of feature-extracting processes.",
    )

    args: argparse.Namespace = parser.parse_args()

    for split in ["id", "ood_ads", "ood_cat", "ood_both"]:
        main(args, split)



from __future__ import annotations

import time
from typing import TYPE_CHECKING, Literal

import ase
from ase import units
from ase.md.langevin import Langevin

from fairchem.core import FAIRChemCalculator
from fairchem.core.calculate import pretrained_mlip
from fairchem.core.components.runner import Runner

if TYPE_CHECKING:
    from fairchem.core.units.mlip_unit.api.inference import InferenceSettings


# A simple example of running Langevin Dynamics with ASE using fairchem + Ray framework
class ASELangevinUMARunner(Runner):
    def __init__(
        self,
        atoms_list: list[ase.Atoms],  # Changed to list of atoms
        model_name: str = "uma-s-1p1",
        settings: InferenceSettings | Literal["default", "other options"] = "default",
        task_name: str = "omat",
        timestep_fs: float = 1.0,
        temp_k: float = 300.0,
        friction_ps_inv: float = 0.001,
        steps_total: int = 1000,
        warmup_steps: int = 10,
        workers: int = 0,
    ):
        self.atoms_list = atoms_list
        self.timestep_fs = timestep_fs
        self.temp_k = temp_k
        self.friction_ps_inv = friction_ps_inv
        self.steps_total = steps_total
        self.warmup_steps = warmup_steps
        self.settings = settings
        self.model_name = model_name
        self.workers = workers
        self.task_name = task_name

    def run(self):
        print(f"Running MD simulations on {len(self.atoms_list)} systems")
        print(f"Steps: warmup={self.warmup_steps}, production={self.steps_total}")
        print("-" * 80)

        results = []

        for i, atoms in enumerate(self.atoms_list):
            predictor = pretrained_mlip.get_predict_unit(
                self.model_name,
                inference_settings=self.settings,
                device="cuda",
                workers=self.workers,
            )
            calc = FAIRChemCalculator(predictor, task_name=self.task_name)
            natoms = len(atoms)
            print(f"\nSystem {i+1}/{len(self.atoms_list)}: {natoms} atoms")

            # Set calculator for this system
            atoms.calc = calc

            # Create dynamics
            dyn = Langevin(
                atoms,
                timestep=self.timestep_fs * units.fs,
                temperature_K=self.temp_k,
                friction=self.friction_ps_inv / units.fs,
            )

            # Warmup
            print(f"  Warming up ({self.warmup_steps} steps)...")
            warmup_start = time.time()
            dyn.run(steps=self.warmup_steps)
            warmup_time = time.time() - warmup_start
            warmup_qps = self.warmup_steps / warmup_time

            # Attach progress reporter for production run
            dyn.attach(
                lambda atoms=atoms, dyn=dyn: print(
                    f"    Step: {dyn.get_number_of_steps()}, "
                    f"E: {atoms.get_potential_energy():.3f} eV"
                ),
                interval=max(self.steps_total // 10, 1),
            )

            # Production run
            print(f"  Production run ({self.steps_total} steps)...")
            start_time = time.time()
            dyn.run(steps=self.steps_total)
            total_time = time.time() - start_time
            production_qps = self.steps_total / total_time

            # Store results
            result = {
                "system_idx": i,
                "natoms": natoms,
                "warmup_qps": warmup_qps,
                "production_qps": production_qps,
                "total_time": total_time,
                "warmup_time": warmup_time,
            }
            results.append(result)

            print(
                f"  Results: warmup_qps={warmup_qps:.2f}, production_qps={production_qps:.2f}"
            )
            del atoms

        # Summary report
        self._print_summary(results)
        return results

    def _print_summary(self, results):
        """Print a summary table of all results"""
        print("\n" + "=" * 80)
        print("PERFORMANCE SUMMARY")
        print("=" * 80)
        print(
            f"{'System':<8} {'Atoms':<8} {'Warmup QPS':<12} {'Prod QPS':<12} {'Total Time':<12}"
        )
        print("-" * 80)

        for result in results:
            print(
                f"{result['system_idx']+1:<8} "
                f"{result['natoms']:<8} "
                f"{result['warmup_qps']:<12.2f} "
                f"{result['production_qps']:<12.2f} "
                f"{result['total_time']:<12.2f}"
            )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        pass

    def load_state(self, checkpoint_location: str | None) -> None:
        pass



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import functools
import logging
import warnings

import numpy as np
import torch
import torch.nn as nn
from matplotlib import pyplot as plt

from fairchem.core.common.registry import registry
from fairchem.core.common.utils import conditional_grad
from fairchem.core.models.base import HeadInterface
from fairchem.core.models.uma.escn_md import eSCNMDBackbone
from fairchem.core.models.uma.nn.mole import (
    MOLEGlobals,
)
from fairchem.core.models.uma.nn.mole_utils import (
    MOLEInterface,
    convert_model_to_MOLE_model,
    model_search_and_replace,
    recursive_replace_all_linear,
    recursive_replace_so2_MOLE,
    replace_linear_with_MOLE,
    replace_MOLE_with_linear,
)

# This will catch the warning despite its C++ origin
# torch.Tensor.index_reduce is in beta
warnings.filterwarnings(
    "ignore",
    message="index_reduce\\(\\) is in beta",
    category=UserWarning,
)


@registry.register_model("escnmd_moe_backbone")
class eSCNMDMoeBackbone(eSCNMDBackbone, MOLEInterface):
    def __init__(
        self,
        num_experts: int = 8,
        moe_dropout: float = 0.0,
        use_global_embedding: bool = False,  # obsolete
        use_composition_embedding: bool = False,
        moe_expert_coefficient_norm: str = "softmax",
        act=torch.nn.SiLU,
        layers_moe=None,
        moe_layer_type: str = "pytorch",
        moe_single: bool = False,
        moe_type: str = "so2",
        model_version: float = 1.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.parent_kwargs = kwargs
        self.num_experts = num_experts
        self.model_version = model_version
        if num_experts > 0:
            convert_model_to_MOLE_model(
                model=self,
                num_experts=num_experts,
                mole_dropout=moe_dropout,
                mole_expert_coefficient_norm=moe_expert_coefficient_norm,
                act=act,
                layers_mole=layers_moe,
                use_composition_embedding=use_composition_embedding,
                mole_layer_type=moe_layer_type,
                mole_single=moe_single,
                mole_type=moe_type,
            )

    def merge_MOLE_model(self, data):
        if self.num_experts == 0:
            return self
        csd_mixed_emb = self.csd_embedding(
            charge=data["charge"],
            spin=data["spin"],
            dataset=data["dataset"],
        )
        self.set_MOLE_coefficients(
            atomic_numbers_full=data["atomic_numbers"],
            batch_full=data["batch"],
            csd_mixed_emb=csd_mixed_emb,
        )
        if self.mole_type != "so2":
            raise ValueError("Only mole_type=so2 supported for merging")

        model_search_and_replace(
            self, recursive_replace_so2_MOLE, replace_MOLE_with_linear
        )

        # drop moe parameters from merged model
        self.routing_mlp = None
        self.composition_embedding = None
        self.num_experts = 0

        # create a new non moe model and load weights into there
        new_model = eSCNMDBackbone(**self.parent_kwargs)
        new_model.load_state_dict(self.state_dict())
        new_model.eval()
        return new_model

    def set_MOLE_coefficients(self, atomic_numbers_full, batch_full, csd_mixed_emb):
        if self.num_experts == 0:
            return
        with torch.autocast(device_type=atomic_numbers_full.device.type, enabled=False):
            embeddings = []
            if self.use_composition_embedding:
                composition_by_atom = self.composition_embedding(atomic_numbers_full)
                composition = composition_by_atom.new_zeros(
                    csd_mixed_emb.shape[0],
                    self.sphere_channels,
                ).index_reduce_(
                    0,
                    batch_full,
                    composition_by_atom,
                    reduce="mean",
                    include_self=np.isclose(self.model_version, 1.0).item(),
                )
                embeddings.append(composition.unsqueeze(0))
            embeddings.append(csd_mixed_emb[None])

            expert_mixing_coefficients_before_norm = self.routing_mlp(
                torch.vstack(embeddings)
                .transpose(0, 1)
                .reshape(csd_mixed_emb.shape[0], -1)
            )
            self.global_mole_tensors.expert_mixing_coefficients = (
                self.mole_expert_coefficient_norm(
                    self.mole_dropout(expert_mixing_coefficients_before_norm)
                )
            )

    def set_MOLE_sizes(self, nsystems, batch_full, edge_index):
        if self.num_experts == 0:
            return
        with torch.autocast(device_type=batch_full.device.type, enabled=False):
            # Generate edge mix_size routing each edge in this instance (GP or not)
            # using its local edge and batch routing

            # Local edge_index is 2xN where [1,:] is the target node, the target node does not
            # have the gp offset applied, which means we need to lookup in the full batch_full
            # _, mix_size = torch.unique(data.batch_full[edge_index[1]], return_counts=True)
            mole_sizes = torch.zeros(
                nsystems,  # data.natoms.shape[0],
                dtype=torch.int,
                device=batch_full[edge_index[1]].device,
            ).scatter_(0, batch_full[edge_index[1]], 1, reduce="add")

            self.global_mole_tensors.mole_sizes = mole_sizes.cpu()

    def log_MOLE_stats(self):
        if not self.training or self.num_experts == 0:
            return
        if not hasattr(self, "fig"):
            self.fig, self.axs = plt.subplots(2, 1)
        with torch.no_grad():
            if self.counter % 500 == 0:
                logging.info(
                    f"{self.counter }: Expert variance: "
                    + ",".join(
                        [
                            f"{x:.2e}"
                            for x in self.global_mole_tensors.expert_mixing_coefficients.var(
                                axis=0
                            ).tolist()
                        ]
                    )
                )
                logging.info(
                    f"{self.counter }: Expert mean: "
                    + ",".join(
                        [
                            f"{x:.2e}"
                            for x in self.global_mole_tensors.expert_mixing_coefficients.mean(
                                axis=0
                            ).tolist()
                        ]
                    )
                )
                self.fig.tight_layout()
                self.plot_ready = True

        self.counter += 1


class DatasetSpecificMoEWrapper(nn.Module, HeadInterface):
    def __init__(
        self,
        backbone,
        dataset_names,
        head_cls,
        wrap_property=True,
        head_kwargs=None,
    ):
        super().__init__()
        if head_kwargs is None:
            head_kwargs = {}
        self.regress_stress = backbone.regress_stress
        self.regress_forces = backbone.regress_forces

        self.wrap_property = wrap_property

        self.dataset_names = sorted(dataset_names)
        self.dataset_name_to_exp = {
            value: idx for idx, value in enumerate(self.dataset_names)
        }
        self.head = registry.get_model_class(head_cls)(backbone, **head_kwargs)
        # replace all linear layers in the head with MOLE
        self.global_mole_tensors = MOLEGlobals(
            expert_mixing_coefficients=None, mole_sizes=None
        )
        replacement_factory = functools.partial(
            replace_linear_with_MOLE,
            num_experts=len(self.dataset_names),
            global_mole_tensors=self.global_mole_tensors,
            mole_layer_type="pytorch",
            cache=None,
        )
        recursive_replace_all_linear(self.head, replacement_factory)

    @conditional_grad(torch.enable_grad())
    def forward(self, data, emb: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        self.global_mole_tensors.mole_sizes = torch.zeros(
            data.natoms.shape[0], dtype=torch.int, device=emb["batch"].device
        ).scatter(0, emb["batch"], 1, reduce="add")  # data.natoms.cpu()
        self.global_mole_tensors.natoms = emb["batch"].shape[0]
        data_batch_full = data.batch_full.cpu()

        # generate a one hot mask based on dataset , one for each system
        self.global_mole_tensors.expert_mixing_coefficients = (
            torch.zeros(
                data.natoms.shape[0],
                len(self.dataset_name_to_exp),
                dtype=data.pos.dtype,
            )
            .scatter(
                1,
                torch.tensor(
                    [
                        self.dataset_name_to_exp[dataset_name]
                        for dataset_name in data.dataset
                    ],
                ).unsqueeze(1),
                1.0,
            )
            .to(data.pos.device)
        )

        # run the internal head
        head_output = self.head(data, emb)

        # breakout the outputs to correct heads named by datasetname
        np_dataset_names = np.array(data.dataset)
        full_output = {}
        for dataset_name in self.dataset_names:
            dataset_mask = np_dataset_names == dataset_name
            for key, mole_output_tensor in head_output.items():
                # TODO cant we use torch.zeros here?
                output_tensor = mole_output_tensor.new_zeros(
                    mole_output_tensor.shape
                )  # float('inf'))
                if dataset_mask.any():
                    if output_tensor.shape[0] == dataset_mask.shape[0]:
                        output_tensor[dataset_mask] = mole_output_tensor[dataset_mask]
                    else:  # assume atoms are the first dimension
                        atoms_mask = torch.isin(
                            data_batch_full,
                            torch.where(torch.from_numpy(dataset_mask))[0],
                        )
                        output_tensor[atoms_mask] = mole_output_tensor[atoms_mask]
                full_output[f"{dataset_name}_{key}"] = (
                    {key: output_tensor} if self.wrap_property else output_tensor
                )
        return full_output


class DatasetSpecificSingleHeadWrapper(nn.Module, HeadInterface):
    def __init__(
        self, backbone, dataset_names, head_cls, wrap_property=True, head_kwargs=None
    ):
        super().__init__()
        if head_kwargs is None:
            head_kwargs = {}
        self.regress_stress = backbone.regress_stress
        self.regress_forces = backbone.regress_forces

        self.wrap_property = wrap_property

        self.dataset_names = sorted(dataset_names)
        self.head = registry.get_model_class(head_cls)(backbone, **head_kwargs)

        # keep track if this head has been merged or not
        self.merged_on_dataset = None

    def merge_MOLE_model(self, data):
        self.merged_on_dataset = data.dataset[0]
        self.non_merged_dataset_names = [
            name for name in self.dataset_names if name != self.merged_on_dataset
        ]
        return self

    def prepare_for_inference(self, data, settings):
        """
        Prepare head for inference. Handles MOLE merging if needed.
        """
        if settings.merge_mole:
            return self.merge_MOLE_model(data)
        return self

    @conditional_grad(torch.enable_grad())
    def forward(self, data, emb: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        data_batch_full = data.batch_full.cpu()
        # run the internal head
        head_output = self.head(data, emb)

        # if merged just return the merged outputs
        # put NaN in other dataset outputs just in case
        # someone tries to use them
        if self.merged_on_dataset is not None:
            full_output = {}
            for key in head_output:
                full_output[f"{self.merged_on_dataset}_{key}"] = (
                    {key: head_output[key]} if self.wrap_property else head_output[key]
                )
                nan_tensor = head_output[key].new_full(
                    head_output[key].shape, float("nan")
                )
                for dataset in self.non_merged_dataset_names:
                    full_output[f"{dataset}_{key}"] = (
                        {key: nan_tensor} if self.wrap_property else nan_tensor
                    )
            return full_output

        # check that all the input dataset names is a strict subset of dataset names
        assert (
            set(data.dataset) <= set(self.dataset_names)
        ), f"Input dataset names: {set(data.dataset)} must be a strict subset of model's valid datset names: {set(self.dataset_names)} "
        # breakout the outputs to correct heads named by datasetname
        np_dataset_names = np.array(data.dataset)

        full_output = {}
        for dataset_name in self.dataset_names:
            dataset_mask = np_dataset_names == dataset_name
            for key, head_output_tensor in head_output.items():
                # TODO cant we use torch.zeros here?
                output_tensor = head_output_tensor.new_zeros(
                    head_output_tensor.shape
                )  # float('inf'))
                if dataset_mask.any():
                    if output_tensor.shape[0] == dataset_mask.shape[0]:
                        output_tensor[dataset_mask] = head_output_tensor[dataset_mask]
                    else:  # assume atoms are the first dimension
                        atoms_mask = torch.isin(
                            data_batch_full,
                            torch.where(torch.from_numpy(dataset_mask))[0],
                        )
                        output_tensor[atoms_mask] = head_output_tensor[atoms_mask]
                full_output[f"{dataset_name}_{key}"] = (
                    {key: output_tensor} if self.wrap_property else output_tensor
                )

        return full_output



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import json
from dataclasses import dataclass
from importlib import resources
from typing import TYPE_CHECKING, Literal

from huggingface_hub import hf_hub_download
from omegaconf import OmegaConf

from fairchem.core import calculate
from fairchem.core._config import CACHE_DIR
from fairchem.core.units.mlip_unit import MLIPPredictUnit, load_predict_unit

if TYPE_CHECKING:
    from fairchem.core.units.mlip_unit import InferenceSettings


@dataclass
class HuggingFaceCheckpoint:
    filename: str
    repo_id: Literal["facebook/UMA"]
    subfolder: str | None = None  # specify a hf repo subfolder
    revision: str | None = None  # specify a version tag, branch, commit hash
    atom_refs: dict | None = None  # specify an isolated atomic reference
    form_elem_refs: dict | None = None  # specify a form elemental reference


@dataclass
class PretrainedModels:
    checkpoints: dict[str, HuggingFaceCheckpoint]


with (resources.files(calculate) / "pretrained_models.json").open("rb") as f:
    _MODEL_CKPTS = PretrainedModels(
        checkpoints={
            model_name: HuggingFaceCheckpoint(**hf_kwargs)
            for model_name, hf_kwargs in json.load(f).items()
        }
    )

available_models = tuple(_MODEL_CKPTS.checkpoints.keys())


def pretrained_checkpoint_path_from_name(model_name: str):
    try:
        model_checkpoint = _MODEL_CKPTS.checkpoints[model_name]
    except KeyError as err:
        raise KeyError(
            f"Model '{model_name}' not found. Available models: {available_models}"
        ) from err
    checkpoint_path = hf_hub_download(
        filename=model_checkpoint.filename,
        repo_id=model_checkpoint.repo_id,
        subfolder=model_checkpoint.subfolder,
        revision=model_checkpoint.revision,
        cache_dir=CACHE_DIR,
    )
    return checkpoint_path


def get_predict_unit(
    model_name: str,
    inference_settings: InferenceSettings | str = "default",
    overrides: dict | None = None,
    device: Literal["cuda", "cpu"] | None = None,
    cache_dir: str = CACHE_DIR,
    workers: int = 1,
) -> MLIPPredictUnit:
    """
    Retrieves a prediction unit for a specified model.

    Args:
        model_name: Name of the model to load from available pretrained models.
        inference_settings: Settings for inference. Can be "default" (general purpose) or "turbo"
            (optimized for speed but requires fixed atomic composition). Advanced use cases can
            use a custom InferenceSettings object.
        overrides: Optional dictionary of settings to override default inference settings.
        device: Optional torch device to load the model onto. If None, uses the default device.
        cache_dir: Path to folder where model files will be stored. Default is "~/.cache/fairchem"
        workers: Number of parallel workers for prediction unit. Default is 1. If greater than 1,
            we will instantiate a ParallelMLIPPredictUnit instead of the normal predict unit.

    Returns:
        An initialized MLIPPredictUnit ready for making predictions.

    Raises:
        KeyError: If the specified model_name is not found in available models.
    """
    checkpoint_path = pretrained_checkpoint_path_from_name(model_name)
    atom_refs = get_reference_energies(model_name, "atom_refs", cache_dir)

    if _MODEL_CKPTS.checkpoints[model_name].form_elem_refs is not None:
        form_elem_refs = get_reference_energies(
            model_name, "form_elem_refs", cache_dir
        )["refs"]
    else:
        form_elem_refs = None

    return load_predict_unit(
        checkpoint_path,
        inference_settings,
        overrides,
        device,
        atom_refs,
        form_elem_refs,
        workers,
    )


def get_reference_energies(
    model_name: str,
    reference_type: Literal["atom_refs", "form_elem_refs"] = "atom_refs",
    cache_dir: str = CACHE_DIR,
) -> dict:
    """
    Retrieves the isolated atomic energies for use with single atom systems into the CACHE_DIR

    Args:
        model_name: Name of the model to load from available pretrained models.
        reference_type: Type of references file to download: atom_refs or bulk_refs.
        cache_dir: Path to folder where files will be stored. Default is "~/.cache/fairchem"
    Returns:
        Atomic or bulk phase element reference data

    Raises:
        KeyError: If the specified model_name is not found in available models.
    """
    model_checkpoint = _MODEL_CKPTS.checkpoints[model_name]
    file_data = getattr(model_checkpoint, reference_type)
    refs_path = hf_hub_download(
        filename=file_data["filename"],
        repo_id=model_checkpoint.repo_id,
        subfolder=file_data["subfolder"],
        revision=model_checkpoint.revision,
        cache_dir=cache_dir,
    )
    return OmegaConf.load(refs_path)



from __future__ import annotations

import argparse
import glob
import logging
import multiprocessing as mp
import os
import random
from pathlib import Path

import numpy as np
from ase.db import connect
from ase.io import read
from tqdm import tqdm

from fairchem.core.datasets import AseDBDataset

logging.basicConfig(level=logging.INFO)


def compute_normalizer_and_linear_reference(train_path, num_workers):
    """
    Given a path to an ASE database file, compute the normalizer value and linear
    reference coefficients. These are used to normalize energies and forces during
    training. For large datasets, compute this for only a subset of the data.
    """
    global dataset
    dataset = AseDBDataset({"src": str(train_path)})

    sample_indices = random.sample(range(len(dataset)), min(100000, len(dataset)))
    with mp.Pool(num_workers) as pool:
        outputs = list(
            tqdm(
                pool.imap(extract_energy_and_forces, sample_indices),
                total=len(sample_indices),
                desc="Computing normalizer values.",
            )
        )
        atomic_numbers = [x[0] for x in outputs]
        energies = [x[1] for x in outputs]
        forces = np.array([force for x in outputs for force in x[2]])
        force_rms = np.sqrt(np.mean(np.square(forces)))
        assert (
            np.isfinite(force_rms).all()
        ), "We found non-finite values in the forces, please check your input data!"
        coeff = compute_lin_ref(atomic_numbers, energies)
    return force_rms, coeff


def extract_energy_and_forces(idx):
    """
    Extract energy and forces from an ASE atoms object at a given index in the dataset.
    """
    atoms = dataset.get_atoms(idx)
    atomic_numbers = atoms.get_atomic_numbers()
    energy = atoms.get_potential_energy()
    forces = atoms.get_forces()
    n_atoms = len(atoms)

    fixed_idx = np.zeros(n_atoms)
    if hasattr(atoms, "constraints"):
        from ase.constraints import FixAtoms

        for constraint in atoms.constraints:
            if isinstance(constraint, FixAtoms):
                fixed_idx[constraint.index] = 1

    mask = fixed_idx == 0
    forces = forces[mask]

    return atomic_numbers, energy, forces


def compute_lin_ref(atomic_numbers, energies):
    """
    Compute linear reference coefficients given atomic numbers and energies.
    """
    features = [np.bincount(x, minlength=100).astype(int) for x in atomic_numbers]

    X = np.vstack(features)
    y = energies

    coeff = np.linalg.lstsq(X, y, rcond=None)[0]
    assert np.isfinite(
        coeff
    ).all(), "Found some non-finite values while computing element references, please check/sanitize your inputs before proceeding"
    return coeff.tolist()


def write_ase_db(mp_arg):
    """
    Write ASE atoms objects to an ASE database file. This function is designed to be
    run in parallel using multiprocessing.
    """
    db_file, file_list, worker_id = mp_arg

    successful = []
    failed = []
    natoms = []
    with connect(str(db_file)) as db:
        for file in tqdm(file_list, position=worker_id):
            atoms_list = read(file, ":")
            for i, atoms in enumerate(atoms_list):
                try:
                    assert (
                        atoms.calc is not None
                    ), "No calculator attached to atoms object."
                    assert "energy" in atoms.calc.results, "Missing energy result"
                    assert "forces" in atoms.calc.results, "Missing forces result"
                    db.write(atoms, data=atoms.info)
                    natoms.append(len(atoms))
                    successful.append(f"{file},{i}")
                except AssertionError as e:
                    failed.append(f"{file},{i}: {e!s}")

    return db_file, natoms, successful, failed


def launch_processing(data_dir, output_dir, num_workers):
    """
    Driver script to launch processing of ASE atoms files into an ASE database.
    """
    os.makedirs(output_dir, exist_ok=True)
    input_files = [
        f
        for f in glob.glob(os.path.join(data_dir, "**/*"), recursive=True)
        if os.path.isfile(f)
    ]
    chunked_files = np.array_split(input_files, num_workers)
    db_files = [output_dir / f"data.{i:04d}.aselmdb" for i in range(num_workers)]
    mp_args = [(db_files[i], chunked_files[i], i) for i in range(num_workers)]
    with mp.Pool(num_workers) as pool:
        outputs = pool.map(write_ase_db, mp_args)

    # Log results
    natoms = []
    for output in outputs:
        db_file, _natoms, successful, failed = output
        natoms.extend(_natoms)
        log_file = db_file.with_suffix(".log")
        failed_file = db_file.with_suffix(".failed")

        with open(log_file, "w") as log:
            log.write("\n".join(successful))
        with open(failed_file, "w") as failed_log:
            failed_log.write("\n".join(failed))
    np.savez_compressed(output_dir / "metadata.npz", natoms=natoms)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--train-dir",
        type=str,
        required=True,
        help="Directory of ASE atoms objects to convert for training.",
    )
    parser.add_argument(
        "--val-dir",
        type=str,
        required=True,
        help="Directory of ASE atoms objects to convert for validation.",
    )
    parser.add_argument(
        "--output-dir",
        type=Path,
        required=True,
        help="Output directory to save required finetuning artifacts.",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=8,
        help="Number of parallel workers for processing files.",
    )
    args = parser.parse_args()

    # Launch processing for training data
    train_path = args.output_dir / "train"
    launch_processing(args.train_dir, train_path, args.num_workers)
    force_rms, linref_coeff = compute_normalizer_and_linear_reference(
        train_path, args.num_workers
    )
    val_path = args.output_dir / "val"
    launch_processing(args.val_dir, val_path, args.num_workers)

    logging.info(f"Generated dataset at {args.output_dir}")



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import torch


def cg_change_mat(ang_mom: int, device: str = "cpu") -> torch.tensor:
    if ang_mom not in [2]:
        raise NotImplementedError

    if ang_mom == 2:
        change_mat = torch.tensor(
            [
                [3 ** (-0.5), 0, 0, 0, 3 ** (-0.5), 0, 0, 0, 3 ** (-0.5)],
                [0, 0, 0, 0, 0, 2 ** (-0.5), 0, -(2 ** (-0.5)), 0],
                [0, 0, -(2 ** (-0.5)), 0, 0, 0, 2 ** (-0.5), 0, 0],
                [0, 2 ** (-0.5), 0, -(2 ** (-0.5)), 0, 0, 0, 0, 0],
                [0, 0, 0.5**0.5, 0, 0, 0, 0.5**0.5, 0, 0],
                [0, 2 ** (-0.5), 0, 2 ** (-0.5), 0, 0, 0, 0, 0],
                [
                    -(6 ** (-0.5)),
                    0,
                    0,
                    0,
                    2 * 6 ** (-0.5),
                    0,
                    0,
                    0,
                    -(6 ** (-0.5)),
                ],
                [0, 0, 0, 0, 0, 2 ** (-0.5), 0, 2 ** (-0.5), 0],
                [-(2 ** (-0.5)), 0, 0, 0, 0, 0, 0, 0, 2 ** (-0.5)],
            ],
            device=device,
        ).detach()

    return change_mat


def irreps_sum(ang_mom: int) -> int:
    """
    Returns the sum of the dimensions of the irreps up to the specified angular momentum.

    :param ang_mom: max angular momenttum to sum up dimensions of irreps
    """
    total = 0
    for i in range(ang_mom + 1):
        total += 2 * i + 1

    return total



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse
import logging
from pathlib import Path

from fairchem.experimental.legacy.utils import load_config

from fairchem.core.common.registry import registry
from fairchem.core.common.utils import save_checkpoint
from fairchem.core.modules.normalization.element_references import (
    create_element_references,
)
from fairchem.core.modules.normalization.normalizer import fit_normalizers


def fit_norms(
    config: dict,
    output_path: str | Path,
    linref_file: str | Path | None = None,
    linref_target: str = "energy",
) -> None:
    """Fit dataset mean and std using the standard config

    Args:
        config: config
        output_path: output path
        linref_file: path to fitted linear references. IF these are used in training they must be used to compute mean/std
        linref_target: target using linear references, basically always energy.
    """
    output_path = Path(output_path).resolve()
    elementrefs = (
        {linref_target: create_element_references(linref_file)}
        if linref_file is not None
        else {}
    )

    try:
        # load the training dataset
        train_dataset = registry.get_dataset_class(
            config["dataset"]["train"].get("format", "lmdb")
        )(config["dataset"]["train"])
    except KeyError as err:
        raise ValueError("Train dataset is not specified in config!") from err

    try:
        norm_config = config["dataset"]["train"]["transforms"]["normalizer"]["fit"]
    except KeyError as err:
        raise ValueError(
            "The provided config does not specify a 'fit' block for 'normalizer'!"
        ) from err

    targets = list(norm_config["targets"].keys())
    override_values = {
        target: vals
        for target, vals in norm_config["targets"].items()
        if isinstance(vals, dict)
    }

    normalizers = fit_normalizers(
        targets=targets,
        override_values=override_values,
        element_references=elementrefs,
        dataset=train_dataset,
        batch_size=norm_config.get("batch_size", 32),
        num_batches=norm_config.get("num_batches"),
        num_workers=config.get("optim", {}).get("num_workers", 16),
    )
    path = save_checkpoint(
        normalizers,
        output_path,
        "normalizers.pt",
    )
    logging.info(f"normalizers have been saved to {path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        required=True,
        type=Path,
        help="Path to configuration yaml file",
    )
    parser.add_argument(
        "--out-path",
        default=".",
        type=str,
        help="Output path to save normalizers",
    )
    parser.add_argument(
        "--linref-path",
        type=str,
        help="Path to linear references used.",
    )
    parser.add_argument(
        "--linref-target",
        default="energy",
        type=str,
        help="target for which linear references are used.",
    )
    args = parser.parse_args()
    config, dup_warning, dup_error = load_config(args.config)

    if len(dup_warning) > 0:
        logging.warning(
            f"The following keys in the given config have duplicates: {dup_warning}."
        )
    if len(dup_error) > 0:
        raise RuntimeError(
            f"The following include entries in the config have duplicates: {dup_error}"
        )

    fit_norms(config, args.out_path, args.linref_path)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import math
import re
import readline
import sys
from itertools import islice
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import torch
import torch.nn as nn
from fairchem.experimental.legacy.utils import build_config
from torch.nn.parallel.distributed import DistributedDataParallel

from fairchem.core.common.flags import flags
from fairchem.core.common.utils import new_trainer_context, setup_logging
from fairchem.core.modules.scaling import ScaleFactor
from fairchem.core.modules.scaling.compat import load_scales_compat

if TYPE_CHECKING:
    from fairchem.core.trainers.base_trainer import BaseTrainer


def _prefilled_input(prompt: str, prefill: str = "") -> str:
    readline.set_startup_hook(lambda: readline.insert_text(prefill))
    try:
        return input(prompt)
    finally:
        readline.set_startup_hook()


def _train_batch(trainer: BaseTrainer, batch) -> None:
    with torch.no_grad():
        with torch.autocast("cuda", enabled=trainer.scaler is not None):
            out = trainer._forward(batch)
        loss = trainer._compute_loss(out, batch)
        del out, loss


def compute_scaling_factors(config, num_batches: int = 16) -> None:
    with new_trainer_context(config=config) as ctx:
        config = ctx.config
        trainer = ctx.trainer

        ckpt_file = config.get("checkpoint", None)
        assert (
            ckpt_file is not None
        ), "Checkpoint file not specified. Please specify --checkpoint <path>"
        ckpt_file = Path(ckpt_file)

        logging.info(f"Input checkpoint path: {ckpt_file}, {ckpt_file.exists()=}")

        model: nn.Module = trainer.model
        data_loader = trainer.train_loader
        assert data_loader is not None, "Train set required to load batches"

        if ckpt_file.exists():
            trainer.load_checkpoint(checkpoint_path=str(ckpt_file))

        # region reoad scale file contents if necessary
        # unwrap module from DP/DDP
        unwrapped_model = model
        while isinstance(unwrapped_model, DistributedDataParallel):
            unwrapped_model = unwrapped_model.module
        assert isinstance(unwrapped_model, nn.Module), "Model is not a nn.Module"
        load_scales_compat(unwrapped_model, config.get("scale_file", None))
        # endregion

        model.eval()

        # recursively go through the submodules and get the ScaleFactor modules
        scale_factors: dict[str, ScaleFactor] = {
            name: module
            for name, module in model.named_modules()
            if isinstance(module, ScaleFactor)
        }

        mode: Literal["all", "unfitted"] = "all"

        # region detect fitted/unfitted factors
        fitted_scale_factors = [
            f"{name}: {module.scale_factor.item():.3f}"
            for name, module in scale_factors.items()
            if module.fitted
        ]
        unfitted_scale_factors = [
            name for name, module in scale_factors.items() if not module.fitted
        ]
        fitted_scale_factors_str = ", ".join(fitted_scale_factors)
        logging.info(f"Fitted scale factors: [{fitted_scale_factors_str}]")
        unfitted_scale_factors_str = ", ".join(unfitted_scale_factors)
        logging.info(f"Unfitted scale factors: [{unfitted_scale_factors_str}]")

        if fitted_scale_factors:
            flag = input(
                "Do you want to continue and fit all scale factors (1), "
                "only fit the variables not fitted yet (2), or exit (3)? "
            )
            if str(flag) == "1":
                mode = "all"
                logging.info("Fitting all scale factors.")
            elif str(flag) == "2":
                mode = "unfitted"
                logging.info("Only fitting unfitted variables.")
            elif str(flag) == "3":
                logging.info("Exiting script")
                sys.exit()
            else:
                logging.error(
                    f"Unrecognized flag associated with fitted_scale_factors: '{flag}'. Exiting."
                )
                sys.exit(-1)
        # endregion

        if ckpt_file.exists():
            logging.warning(f"Already found existing file: {ckpt_file}")
            flag = input(
                "Do you want to continue and overwrite existing file (1), "
                "or exit (2)? "
            )
            if str(flag) == "1":
                logging.info("Overwriting existing file.")
            else:
                logging.info("Exiting script")
                sys.exit()

        logging.info(
            f"Output path for fitted scale factors: {ckpt_file}, {ckpt_file.exists()=}"
        )
        # endregion

        # region reset the scale factors if mode == "all"
        if mode == "all":
            logging.info("Fitting all scale factors.")
            for name, scale_factor in scale_factors.items():
                if scale_factor.fitted:
                    logging.info(
                        f"{name} is already fitted in the checkpoint, resetting it. {scale_factor.scale_factor}"
                    )
                scale_factor.reset_()
        # endregion

        # region we do a single pass through the network to get the correct execution order of the scale factors
        scale_factor_indices: dict[str, int] = {}
        max_idx = 0

        # initialize all scale factors
        for name, module in scale_factors.items():

            def index_fn(name: str = name) -> None:
                nonlocal max_idx
                assert name is not None
                if name not in scale_factor_indices:
                    scale_factor_indices[name] = max_idx
                    logging.debug(f"Scale factor for {name} = {max_idx}")
                    max_idx += 1

            module.initialize_(index_fn=index_fn)

        # single pass through network
        _train_batch(trainer, next(iter(data_loader)))

        # sort the scale factors by their computation order
        sorted_factors = sorted(
            scale_factors.items(),
            key=lambda x: scale_factor_indices.get(x[0], math.inf),
        )

        logging.info("Sorted scale factors by computation order:")
        for name, _ in sorted_factors:
            logging.info(f"{name}: {scale_factor_indices[name]}")

        # endregion

        # loop over the scale factors in the computation order
        # and fit them one by one
        logging.info("Start fitting")

        for name, module in sorted_factors:
            if mode == "unfitted" and module.fitted:
                logging.info(f"Skipping {name} (already fitted)")
                continue

            logging.info(f"Fitting {name}...")
            with module.fit_context_():
                for batch in islice(data_loader, num_batches):
                    _train_batch(trainer, batch)
                stats, ratio, value = module.fit_()

                logging.info(
                    f"Variable: {name}, "
                    f"Var_in: {stats['variance_in']:.3f}, "
                    f"Var_out: {stats['variance_out']:.3f}, "
                    f"Ratio: {ratio:.3f} => Scaling factor: {value:.3f}"
                )

        # make sure all scale factors are fitted
        for name, module in sorted_factors:
            assert module.fitted, f"{name} is not fitted"

        # region save the scale factors to the checkpoint file
        trainer.config["cmd"]["checkpoint_dir"] = ckpt_file.parent
        trainer.is_debug = False

        def rename_module(name):
            name = name.replace(".scale_factor", "")
            # remove DDP wrapper
            name = re.sub("^module.", "", name)
            # remove hydra backbone
            return re.sub("^backbone.", "", name)

        torch.save(
            {
                rename_module(x[0]): x[1]
                for x in trainer.model.to("cpu").named_parameters()
                if ".scale_" in x[0]
            },
            str(ckpt_file),
        )
        logging.info(f"Saved results to: {ckpt_file}")


if __name__ == "__main__":
    # region args/config setup
    setup_logging()

    parser = flags.get_parser()
    args, override_args = parser.parse_known_args()
    assert not args.distributed, "This doesn't work with DDP"
    config = build_config(args, override_args)

    compute_scaling_factors(config)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from .adsorbml_runner import AdsorbMLRunner
from .adsorption_runner import AdsorptionRunner
from .adsorption_singlepoint_runner import AdsorptionSinglePointRunner
from .elasticity_runner import ElasticityRunner
from .kappa_runner import KappaRunner
from .nve_md_runner import NVEMDRunner
from .omol_runner import OMolRunner
from .pairwise_ct_runner import PairwiseCountRunner
from .phonon_runner import MDRPhononRunner
from .relaxation_runner import RelaxationRunner
from .singlepoint_runner import SinglePointRunner

__all__ = [
    "AdsorptionRunner",
    "AdsorbMLRunner",
    "AdsorptionSinglePointRunner",
    "ElasticityRunner",
    "KappaRunner",
    "NVEMDRunner",
    "OMolRunner",
    "PairwiseCountRunner",
    "MDRPhononRunner",
    "RelaxationRunner",
    "SinglePointRunner",
]
